# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""

import os

import numpy as np
import pandas as pd
import xarray as xr
from pyproj import CRS
from siphon.catalog import TDSCatalog
from xarray.backends import NetCDF4DataStore


class Dataset:
    def __init__(self):
        self.quantity = None
        self.unit = None
        self.time = []
        self.x = None
        self.y = None
        self.crs = None
        self.val = None


#        self.u        = None
#        self.v        = None


def download(
    param_list, lon_range, lat_range, path, prefix, time_range=None, times=None
):
    base_url = "https://www.ncei.noaa.gov/thredds/catalog/model-gfs-g4-anl-files-old/"

    if times is not None:
        requested_times = times
        time_range = [times[0], times[-1]]
    else:
        requested_times = (
            pd.date_range(start=time_range[0], end=time_range[1], freq="3H")
            .to_pydatetime()
            .tolist()
        )

    ntime = len(requested_times)

    datasets = {}

    for param in param_list:
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)
        dataset.quantity = param
        datasets[param] = dataset

    icont = False
    # Get lat,lon
    for it, time in enumerate(requested_times):
        h = requested_times[it].hour
        month_string = requested_times[it].strftime("%Y%m")
        date_string = requested_times[it].strftime("%Y%m%d")
        url = base_url + month_string + "/" + date_string + "/catalog.xml"
        try:
            gfs = TDSCatalog(url)
        except Exception:
            gfs = []
        cstr = "0000_000"
        name = "gfsanl_4_" + date_string + "_" + cstr + ".grb2"
        okay = False
        if gfs:
            for j, ds in enumerate(gfs.datasets):
                if ds == name:
                    okay = True
                    break
        if not okay:
            # Try the next time
            continue
        ncss = gfs.datasets[j].subset()
        query = ncss.query()
        query.lonlat_box(
            north=lat_range[1], south=lat_range[0], east=lon_range[1], west=lon_range[0]
        ).vertical_level(10.0)
        query.variables("u-component_of_wind_height_above_ground")
        data = ncss.get_data(query)
        data = xr.open_dataset(NetCDF4DataStore(data))
        lon = np.array(data["lon"])
        lat = np.array(data["lat"])
        nrows = len(lat)
        ncols = len(lon)
        # Latitude and longitude found, so we can stop now
        icont = True
        break

    if not icont:
        # Could not find any data
        print("Could not find any data in requested range !")
        datasets = {}
        return datasets

    # initialize matrices
    for param in param_list:
        datasets[param].x = lon
        datasets[param].y = lat
        datasets[param].val = np.empty((ntime, nrows, ncols))
        datasets[param].val[:] = np.NaN

    for it, time in enumerate(requested_times):
        h = time.hour
        month_string = time.strftime("%Y%m")
        date_string = time.strftime("%Y%m%d")
        url = base_url + month_string + "/" + date_string + "/catalog.xml"

        try:
            gfs = TDSCatalog(url)
        except Exception:
            print("Could not fetch catalogue")
            continue

        if h == 0:
            cstr = "0000_000"
            crstr = "0000_003"
            var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
        elif h == 3:
            cstr = "0000_003"
            crstr = "0000_006"
            var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
        elif h == 6:
            cstr = "0600_000"
            crstr = "0600_003"
            var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
        elif h == 9:
            cstr = "0600_003"
            crstr = "0600_006"
            var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
        elif h == 12:
            cstr = "1200_000"
            crstr = "1200_003"
            var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
        elif h == 15:
            cstr = "1200_003"
            crstr = "1200_006"
            var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
        elif h == 18:
            cstr = "1800_000"
            crstr = "1800_003"
            var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
        elif h == 21:
            cstr = "1800_003"
            crstr = "1800_006"
            var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"

        # Loop through requested parameters
        for ind, param in enumerate(param_list):
            datasets[param].time.append(time)

            if param == "precipitation":
                name = "gfsanl_4_" + date_string + "_" + crstr + ".grb2"
            else:
                name = "gfsanl_4_" + date_string + "_" + cstr + ".grb2"

            try:
                okay = False
                for j, ds in enumerate(gfs.datasets):
                    if ds == name:
                        okay = True
                        break

                if not okay:
                    # File not found, on to the next parameter
                    print(name + " was not found on server ...")
                    continue

                print(name + " : " + param)

                ncss = gfs.datasets[j].subset()

                if param == "wind_u":
                    var_name = "u-component_of_wind_height_above_ground"
                elif param == "wind_v":
                    var_name = "v-component_of_wind_height_above_ground"
                elif param == "barometric_pressure":
                    var_name = "Pressure_reduced_to_MSL_msl"
                elif param == "precipitation":
                    var_name = var_prcp

                query = ncss.query()

                if param == "wind_u" or param == "wind_v":
                    query.lonlat_box(
                        north=lat_range[1],
                        south=lat_range[0],
                        east=lon_range[1],
                        west=lon_range[0],
                    ).vertical_level(10.0)
                else:
                    query.lonlat_box(
                        north=lat_range[1],
                        south=lat_range[0],
                        east=lon_range[1],
                        west=lon_range[0],
                    )

                query.variables(var_name)
                data = ncss.get_data(query)
                data = xr.open_dataset(NetCDF4DataStore(data))
                val = data[var_name]
                datasets[param].unit = val.units
                val = np.array(val.metpy.unit_array.squeeze())

                if param == "precipitation":
                    # Data is stored either as 3-hourly (at 03h) or 6-hourly (at 06h) accumulated rainfall
                    # For the first, just divide by 3 to get hourly precip
                    # For the second, first subtract volume that fell in the first 3 hours
                    if h == 0 or h == 6 or h == 12 or h == 18:
                        val = val / 3  # Convert to mm/h
                    else:
                        val = (val - 3 * np.squeeze(dataset.val[it - 1, :, :])) / 3

                datasets[param].val[it, :, :] = val

            except Exception:
                print("Could not download data")

        # Write data to netcdf
        time_string = time.strftime("%Y%m%d_%H%M")
        file_name = prefix + "." + time_string + ".nc"
        full_file_name = os.path.join(path, file_name)
        ds = xr.Dataset()

        okay = False

        for param in param_list:
            val = datasets[param].val[it, :, :]

            if not np.any(np.isnan(val)):
                okay = True
                da = xr.DataArray(
                    val, coords=[("lat", datasets[param].y), ("lon", datasets[param].x)]
                )
                ds[param] = da

        if okay:
            # Only write to file if there is any data
            ds.to_netcdf(path=full_file_name)

    return datasets


# Helper function for finding proper time variable
def find_time_var(var, time_basename="time"):
    for coord_name in var.coords:
        if coord_name.startswith(time_basename):
            return var.coords[coord_name]
    raise ValueError("No time variable found for " + var.name)



# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""

import netCDF4
import numpy as np
import pandas as pd
from pyproj import CRS


class Dataset:
    def __init__(self):
        self.quantity = None
        self.unit = None
        self.time = None
        self.x = None
        self.y = None
        self.crs = None
        self.val = None


def download(param_list, lon_range, lat_range, time_range, cycle_time):
    cycle_string = cycle_time.strftime("%Y%m%d%H")

    base_url = "https://icoast.rc.ufl.edu/thredds/dodsC/coamps/"

    times = (
        pd.date_range(start=time_range[0], end=time_range[1], freq="1H")
        .to_pydatetime()
        .tolist()
    )

    datasets = []

    # Get lon and lat from first time
    nc_file = "coamps-tc_d01_" + cycle_string + "_tau000.nc"
    try:
        nc4 = netCDF4.Dataset(base_url + nc_file)
    except Exception:
        print("Not available: " + base_url + nc_file)
        return datasets

    lon0 = np.squeeze(nc4["lon"][0, :].astype("float64")) - 360.0
    lat0 = np.squeeze(nc4["lat"][:, 0].astype("float64"))

    ilon0 = np.where(lon0 < lon_range[0])[0][-1]
    ilon1 = np.where(lon0 > lon_range[1])[0][0] + 1
    ilat0 = np.where(lat0 < lat_range[0])[0][-1]
    ilat1 = np.where(lat0 > lat_range[1])[0][0] + 1
    lon = lon0[ilon0:ilon1]
    lat = np.flipud(lat0[ilat0:ilat1])
    #    lat = lat0[ilat0:ilat1]

    # Loop through requested parameters
    for param in param_list:
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)

        if param == "wind":
            dataset.time = times
            dataset.x = lon
            dataset.y = lat
            dataset.quantity = param
            dataset.unit = "m/s"
            dataset.u = np.zeros((np.size(times), np.size(lat), np.size(lon)))
            dataset.v = np.zeros((np.size(times), np.size(lat), np.size(lon)))

            iok = -1
            for it, time in enumerate(times):
                hrs = int(
                    (
                        time.replace(tzinfo=None) - cycle_time.replace(tzinfo=None)
                    ).total_seconds()
                    / 3600
                )
                nc_file = (
                    "coamps-tc_d01_" + cycle_string + "_tau" + str(hrs).zfill(3) + ".nc"
                )
                try:
                    nc4 = netCDF4.Dataset(base_url + nc_file)
                    dataset.u[it, :, :] = np.flipud(
                        nc4["uuwind"][ilat0:ilat1, ilon0:ilon1]
                    )
                    dataset.v[it, :, :] = np.flipud(
                        nc4["vvwind"][ilat0:ilat1, ilon0:ilon1]
                    )
                    iok = it

                except Exception:
                    pass
            if iok < np.size(times) - 1:
                dataset.time = dataset.time[0 : iok + 1]
                dataset.u = dataset.u[0 : iok + 1, :, :]
                dataset.v = dataset.v[0 : iok + 1, :, :]

        else:
            # Other scalar variables

            fac = 1.0

            dataset.quantity = param

            if param == "barometric_pressure":
                var_name = "slpres"
                dataset.unit = "Pa"
                fac = 100.0
            elif param == "precipitation":
                var_name = None
                dataset.unit = "mm/h"
                fac = 3600.0

            dataset.time = times
            dataset.x = lon
            dataset.y = lat
            dataset.quantity = param
            dataset.val = np.zeros((np.size(times), np.size(lat), np.size(lon)))

            if var_name:
                iok = -1
                for it, time in enumerate(times):
                    hrs = int(
                        (
                            time.replace(tzinfo=None) - cycle_time.replace(tzinfo=None)
                        ).total_seconds()
                        / 3600
                    )
                    nc_file = (
                        "coamps-tc_d01_"
                        + cycle_string
                        + "_tau"
                        + str(hrs).zfill(3)
                        + ".nc"
                    )
                    try:
                        nc4 = netCDF4.Dataset(base_url + nc_file)
                        dataset.val[it, :, :] = fac * np.flipud(
                            nc4[var_name][ilat0:ilat1, ilon0:ilon1]
                        )
                        iok = it
                    except Exception:
                        pass
                if iok < np.size(times) - 1:
                    dataset.time = dataset.time[0 : iok + 1]
                    dataset.val = dataset.val[0 : iok + 1, :, :]
        if iok >= 0:
            datasets.append(dataset)

    return datasets



import os
import numpy as np

def write_to_delft3d_ascii(
    dataset,
    file_name,
    version="1.03",
    path=None,
    header_comments=False,
    refdate=None,
    parameters=None,
    time_range=None,
):
        
    # Convert numpy.datetime64 to datetime
    time = dataset.ds.time.values.astype("M8[s]").astype("O")

    if not refdate:
        refdate = time[0]

    if not time_range:
        time_range = [time[0], time[-1]]

    if not parameters:
        parameters = ["wind", "barometric_pressure", "precipitation"]

    if dataset.crs.is_geographic:
        grid_unit = "degrees"
    else:
        grid_unit = "m"

    files = []
    for param in parameters:
        if param == "wind":
            file = {}
            file["data"] = dataset.ds["wind_u"].values[:]
            file["ext"] = "amu"
            file["quantity"] = "x_wind"
            file["unit"] = "m s-1"
            file["fmt"] = "%6.1f"
            files.append(file)
            file = {}
            file["data"] = dataset.ds["wind_v"].values[:]
            file["ext"] = "amv"
            file["quantity"] = "y_wind"
            file["unit"] = "m s-1"
            file["fmt"] = "%6.1f"
            files.append(file)
        elif param == "barometric_pressure":
            file = {}
            file["data"] = dataset.ds["barometric_pressure"].values[:]
            file["ext"] = "amp"
            file["quantity"] = "air_pressure"
            file["unit"] = "Pa"
            file["fmt"] = "%7.0f"
            files.append(file)
        elif param == "precipitation":
            file = {}
            file["data"] = dataset.ds["precipitation"].values[:]
            file["ext"] = "ampr"
            file["quantity"] = "precipitation"
            file["unit"] = "mm h-1"
            file["fmt"] = "%7.1f"
            files.append(file)

    # if dataset.quantity == "x_wind":
    #     unit = "m s-1"
    #     ext  = "amu"
    #     fmt  = "%6.1f"
    # elif dataset.quantity == "y_wind":
    #     unit = "m s-1"
    #     ext  = "amv"
    #     fmt  = "%6.1f"
    # elif dataset.quantity == "air_pressure":
    #     unit = "Pa"
    #     ext  = "amp"
    #     fmt  = "%7.0f"
    # elif dataset.quantity == "air_temperature":
    #     unit = "Celsius"
    #     ext  = "amt"
    #     fmt  = "%7.1f"
    # elif dataset.quantity == "relative_humidity":
    #     unit = "%"
    #     ext  = "amr"
    #     fmt  = "%7.1f"
    # elif dataset.quantity == "cloudiness":
    #     unit = "%"
    #     ext  = "amc"
    #     fmt  = "%7.1f"
    # elif dataset.quantity == "sw_radiation_flux":
    #     unit = "W/m2"
    #     ext  = "ams"
    #     fmt  = "%7.1f"
    # elif dataset.quantity == "precipitation":
    #     unit = "mm/h"
    #     ext  = "ampr"
    #     fmt  = "%7.1f"

    for file in files:

        if "lon" in dataset.ds:
            ncols = len(dataset.ds.lon)
            x = dataset.ds.lon.values[:]
        else:
            ncols = len(dataset.ds.x)
            x = dataset.ds.x.values[:]
        if "lat" in dataset.ds:
            nrows = len(dataset.ds.lat)
            y = dataset.ds.lat.values[:]
        else:
            nrows = len(dataset.ds.y)
            y = dataset.ds.y.values[:]

        dx = (x[-1] - x[0]) / (ncols - 1)
        dy = (y[-1] - y[0]) / (nrows - 1)

        if path:
            full_file_name = os.path.join(path, file_name + "." + file["ext"])
        else:
            full_file_name = file_name + "." + file["ext"]

        fid = open(full_file_name, "w")

        if header_comments:
            fid.write("### START OF HEADER\n")
            fid.write(
                "### All text on a line behind the first # is parsed as commentary\n"
            )
            fid.write("### Additional comments\n")

        fid.write(
            "FileVersion      =   "
            + version
            + "                                               # Version of meteo input file, to check if the newest file format is used\n"
        )
        fid.write(
            "filetype         =   meteo_on_equidistant_grid                          # Type of meteo input file: meteo_on_flow_grid, meteo_on_equidistant_grid, meteo_on_curvilinear_grid or meteo_on_spiderweb_grid\n"
        )
        fid.write(
            "NODATA_value     =   -999                                               # Value used for undefined or missing data\n"
        )
        fid.write("n_cols           =   " + str(ncols) + "\n")
        fid.write("n_rows           =   " + str(nrows) + "\n")
        fid.write("grid_unit        =   " + grid_unit + "\n")
        #            fid.write("x_llcorner       =   " + str(min(dataset.x)) + "\n")
        #            fid.write("y_llcorner       =   " + str(min(dataset.y)) + "\n")
        fid.write("x_llcorner       =   " + str(min(x) - 0.5 * dx) + "\n")
        fid.write("y_llcorner       =   " + str(min(y) - 0.5 * dy) + "\n")
        if version == "1.02":
            fid.write("value_pos       =    corner\n")
        # Use max 4 decimals for dx and dy
        if dataset.crs.is_geographic:
            fid.write("dx               =   " + "{:.5f}".format(dx) + "\n")
            fid.write("dy               =   " + "{:.5f}".format(dy) + "\n")
        else:
            fid.write("dx               =   " + "{:.1f}".format(dx) + "\n")
            fid.write("dy               =   " + "{:.1f}".format(dy) + "\n")
        fid.write(
            "n_quantity       =   1                                                  # Number of quantities prescribed in the file\n"
        )
        fid.write("quantity1        =   " + file["quantity"] + "\n")
        fid.write("unit1            =   " + file["unit"] + "\n")
        if header_comments:
            fid.write("### END OF HEADER\n")

        # Add extra blocks if data does not cover time range
        if time[0] > time_range[0]:
            dt = time_range[0] - refdate
            tim = dt.total_seconds() / 60
            val = np.flipud(file["data"][0, :, :])
            # Skip blocks with only nans
            if not np.all(np.isnan(val)):
                val[val == np.nan] = -999.0
                fid.write(
                    "TIME = "
                    + str(tim)
                    + " minutes since "
                    + refdate.strftime("%Y-%m-%d %H:%M:%S")
                    + " +00:00\n"
                )
                np.savetxt(fid, val, fmt=file["fmt"])

        for it, t in enumerate(time):

            dt = t - refdate
            tim = dt.total_seconds() / 60
            val = np.flipud(file["data"][it, :, :])

            if file["quantity"] == "x_wind" or file["quantity"] == "y_wind":
                if np.max(val) > 1000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    val[np.where(val == 0.0)] = np.nan
                    print(
                        "Warning! Wind speed > 1000 m/s at "
                        + t.strftime("%Y-%m-%d %H:%M:%S")
                        + " !"
                    )
                if np.min(val) < -1000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    print(
                        "Warning! Wind speed > 1000 m/s at "
                        + t.strftime("%Y-%m-%d %H:%M:%S")
                        + " !"
                    )
                    val[np.where(val == 0.0)] = np.nan
            if file["quantity"] == "air_pressure":        
                if np.max(val) > 200000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    val[np.where(val == 0.0)] = np.nan
                if np.min(val) < 10000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    val[np.where(val == 0.0)] = np.nan
            if file["quantity"] == "precipitation":
                if np.nanmax(val) > 1000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    print(
                        "Warning! Precipitation exceeds 1000 mm/h at "
                        + t.strftime("%Y-%m-%d %H:%M:%S")
                        + " !"
                    )
                    val[np.where(val == 0.0)] = np.nan
                if np.nanmin(val) < 0.0:
                    val[np.where(val < 0.0)] = 0.0

            if np.all(np.isnan(val)):
                if it > 0:
                    print(
                        "Warning! Only NaNs found for "
                        + param
                        + " at "
                        + t.strftime("%Y-%m-%d %H:%M:%S")
                        + " ! Using data from previous time."
                    )
                    val = val0

                else:
                    if file["quantity"] == "x_wind" or file["quantity"] == "y_wind" or file["quantity"] == "precipitation":
                        print(
                            "Warning! Only NaNs found for "
                            + param
                            + " at "
                            + t.strftime("%Y-%m-%d %H:%M:%S")
                            + " ! Setting values to 0.0 !"
                        )
                        val = np.zeros_like(val)
                    elif file["quantity"] == "air_pressure":
                        print(
                            "Warning! Only NaNs found for "
                            + param
                            + " at "
                            + t.strftime("%Y-%m-%d %H:%M:%S")
                            + " ! Setting values to 101300.0 !"
                        )
                        val = np.zeros_like(val) + 101300.0

            fid.write(
                "TIME = "
                + str(tim)
                + " minutes since "
                + refdate.strftime("%Y-%m-%d %H:%M:%S")
                + " +00:00\n"
            )
            np.savetxt(fid, val, fmt=file["fmt"])
            val0 = val # Saved in case next time only has NaNs

        # Add extra blocks if data does not cover time range
        if time[-1] < time_range[1]:
            dt = time_range[1] - refdate
            tim = dt.total_seconds() / 60
            val = np.flipud(file["data"][-1, :, :])
            # Skip blocks with only nans
            if not np.all(np.isnan(val)):
                val[val == np.nan] = -999.0
                fid.write(
                    "TIME = "
                    + str(tim)
                    + " minutes since "
                    + refdate.strftime("%Y-%m-%d %H:%M:%S")
                    + " +00:00\n"
                )
                np.savetxt(fid, val, fmt=file["fmt"])

        fid.close()




# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""

from siphon.catalog import TDSCatalog
from xarray.backends import NetCDF4DataStore
import xarray as xr
import pandas as pd
import numpy as np
from pyproj import CRS
from metpy.units import units
import netCDF4

class Dataset():

    def __init__(self):

        self.quantity = None
        self.unit     = None
        self.time     = None
        self.x        = None
        self.y        = None
        self.crs      = None
        self.val      = None

def download(param_list, lon_range, lat_range, time_range, cycle_time):

    cycle_string = cycle_time.strftime("%Y%m%d%H")

    base_url = "https://icoast.rc.ufl.edu/thredds/dodsC/coamps/"

    times = pd.date_range(start=time_range[0],
                          end=time_range[1],
                          freq='1H').to_pydatetime().tolist()

    datasets = []

    # Get lon and lat from first time
    nc_file = "coamps-tc_d01_" + cycle_string + "_tau000.nc"
    try:        
        nc4 = netCDF4.Dataset(base_url + nc_file)
    except:
        print("Not available: " + base_url + nc_file)
        return datasets
        
    lon0 = np.squeeze(nc4["lon"][0,:].astype('float64')) - 360.0
    lat0 = np.squeeze(nc4["lat"][:,0].astype('float64'))

    ilon0 = np.where(lon0<lon_range[0])[0][-1]
    ilon1 = np.where(lon0>lon_range[1])[0][0] + 1
    ilat0 = np.where(lat0<lat_range[0])[0][-1]
    ilat1 = np.where(lat0>lat_range[1])[0][0] + 1
    lon = lon0[ilon0:ilon1]
    lat = np.flipud(lat0[ilat0:ilat1])    
#    lat = lat0[ilat0:ilat1]    

    # Loop through requested parameters     
    for param in param_list:
        
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)
 
        if param == "wind":
            
            dataset.time     = times
            dataset.x        = lon
            dataset.y        = lat
            dataset.quantity = param
            dataset.unit     = "m/s"
            dataset.u        = np.zeros((np.size(times), np.size(lat), np.size(lon)))
            dataset.v        = np.zeros((np.size(times), np.size(lat), np.size(lon)))

            iok = -1            
            for it, time in enumerate(times):
                hrs = int((time.replace(tzinfo=None) - cycle_time.replace(tzinfo=None)).total_seconds()/3600)
                nc_file = "coamps-tc_d01_" + cycle_string + "_tau" + str(hrs).zfill(3) + ".nc"
                try:
                    nc4 = netCDF4.Dataset(base_url + nc_file)
                    dataset.u[it,:,:] = np.flipud(nc4["uuwind"][ilat0:ilat1, ilon0:ilon1])
                    dataset.v[it,:,:] = np.flipud(nc4["vvwind"][ilat0:ilat1, ilon0:ilon1])
                    iok = it

                except:
                    pass
            if iok<np.size(times) - 1:
                dataset.time = dataset.time[0:iok + 1]
                dataset.u    = dataset.u[0:iok + 1, :, :]
                dataset.v    = dataset.v[0:iok + 1, :, :]
        
        else:
            
            # Other scalar variables
            
            fac = 1.0
            
            dataset.quantity = param

            if param == "barometric_pressure":
                var_name = "slpres"
                dataset.unit     = "Pa"
                fac = 100.0
            elif param == "precipitation":    
                var_name = None
                dataset.unit     = "mm/h"
                fac = 3600.0

            dataset.time     = times
            dataset.x        = lon
            dataset.y        = lat
            dataset.quantity = param
            dataset.val      = np.zeros((np.size(times), np.size(lat), np.size(lon)))

            if var_name:                
                iok = -1            
                for it, time in enumerate(times):    
                    hrs = int((time.replace(tzinfo=None) - cycle_time.replace(tzinfo=None)).total_seconds()/3600)
                    nc_file = "coamps-tc_d01_" + cycle_string + "_tau" + str(hrs).zfill(3) + ".nc"
                    try:
                        nc4 = netCDF4.Dataset(base_url + nc_file)
                        dataset.val[it,:,:] = fac*np.flipud(nc4[var_name][ilat0:ilat1, ilon0:ilon1])
                        iok = it
                    except:
                        pass
                if iok<np.size(times) - 1:
                    dataset.time = dataset.time[0:iok + 1]
                    dataset.val  = dataset.val[0:iok + 1, :, :]
        if iok>=0:
            datasets.append(dataset)    
    
    return datasets



# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""

import os
from siphon.catalog import TDSCatalog
from xarray.backends import NetCDF4DataStore
import xarray as xr
import pandas as pd
import numpy as np
from pyproj import CRS
from metpy.units import units

class Dataset():

    def __init__(self):

        self.quantity = None
        self.unit     = None
        self.time     = []
        self.x        = None
        self.y        = None
        self.crs      = None
        self.val      = None
        self.u        = None
        self.v        = None

def download(param_list, lon_range, lat_range, path, prefix,
             time_range=None, times=None):

    pass

#     base_url = "https://www.ncei.noaa.gov/thredds/catalog/model-gfs-g4-anl-files-old/"
    

#     if times is not None:
#         requested_times = times
#         time_range = [times[0], times[-1]]
#     else:    
#         requested_times = pd.date_range(start=time_range[0],
#                           end=time_range[1],
#                           freq='3H').to_pydatetime().tolist()

#     ntime = len(requested_times)

#     datasets = []
#     for param in param_list:        
#         dataset = Dataset()
#         dataset.crs = CRS.from_epsg(4326)
#         dataset.quantity = param
#         datasets.append(dataset)

#     icont = False
#     # Get lat,lon
#     for it, time in enumerate(requested_times):
#         h            = requested_times[it].hour
#         month_string = requested_times[it].strftime("%Y%m")
#         date_string  = requested_times[it].strftime("%Y%m%d")
#         url = base_url + month_string + "/" + date_string + "/catalog.xml"
#         try:
#             gfs   = TDSCatalog(url)
#         except:
#             gfs   = []
#         cstr  = "0000_000"
#         name = "gfsanl_4_" + date_string + "_" + cstr + ".grb2"
#         okay = False               
#         if gfs:
#             for j, ds in enumerate(gfs.datasets):
#                 if ds==name:
#                     okay = True
#                     break
#         if not okay:
#             # Try the next time
#             continue
#         ncss  = gfs.datasets[j].subset()                
#         query = ncss.query()            
#         query.lonlat_box(north=lat_range[1], \
#                          south=lat_range[0], \
#                           east=lon_range[1],  \
#                           west=lon_range[0]).vertical_level(10.0)
#         query.variables("u-component_of_wind_height_above_ground",
#                         "v-component_of_wind_height_above_ground")
#         data = ncss.get_data(query)
#         data = xr.open_dataset(NetCDF4DataStore(data))        
#         lon = np.array(data['lon'])
#         lat = np.array(data['lat'])  
#         nrows = len(lat)
#         ncols = len(lon)
#         # Latitude and longitude found, so we can stop now
#         icont = True
#         break
    
#     if not icont:
#         # Could not find any data
#         print("Could not find any data in requested range !")
#         datasets = []
#         return datasets
    
#     # initialize matrices
#     for dataset in datasets:        
#         dataset.x = lon
#         dataset.y = lat
#         if dataset.quantity == "wind":                    
#             dataset.u    = np.empty((ntime, nrows, ncols))
#             dataset.u[:] = np.NaN            
#             dataset.v    = np.empty((ntime, nrows, ncols))                
#             dataset.v[:] = np.NaN            
#         else:
#             dataset.val    = np.empty((ntime, nrows, ncols))
#             dataset.val[:] = np.NaN            
    
#     for it, time in enumerate(requested_times):
        
#         h            = time.hour
#         month_string = time.strftime("%Y%m")
#         date_string  = time.strftime("%Y%m%d")
#         url = base_url + month_string + "/" + date_string + "/catalog.xml"
        
#         try:
#             gfs   = TDSCatalog(url)
#         except:
#             print("Could not fetch catalogue")
#             continue
        
#         if h==0:
#             cstr  = "0000_000"
#             crstr = "0000_003"
#             var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
#         elif h==3:    
#             cstr  = "0000_003"
#             crstr = "0000_006"
#             var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
#         elif h==6:    
#             cstr  = "0600_000"
#             crstr = "0600_003"
#             var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
#         elif h==9:    
#             cstr  = "0600_003"
#             crstr = "0600_006"
#             var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
#         elif h==12:    
#             cstr  = "1200_000"
#             crstr = "1200_003"
#             var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
#         elif h==15:    
#             cstr  = "1200_003"
#             crstr = "1200_006"
#             var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
#         elif h==18:    
#             cstr  = "1800_000"
#             crstr = "1800_003"
#             var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
#         elif h==21:    
#             cstr  = "1800_003"
#             crstr = "1800_006"
#             var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"

#         # Loop through requested parameters
#         for ind, param in enumerate(param_list):
            
#             dataset = datasets[ind]
#             dataset.time.append(time)

#             if param == "precipitation":
#                 name = "gfsanl_4_" + date_string + "_" + crstr + ".grb2" 
#             else:    
#                 name = "gfsanl_4_" + date_string + "_" + cstr + ".grb2" 

#             try:
                            
#                 okay = False
#                 for j, ds in enumerate(gfs.datasets):
#                     if ds==name:
#                         okay = True
#                         break
                
#                 if not okay:                    
#                     # File not found, on to the next parameter
#                     print(name + " was not found on server ...")
#                     continue
    
#                 print(name + " : " + param)    
    
#                 ncss  = gfs.datasets[j].subset()
                
#                 if param == "wind":                    
#                     query = ncss.query()            
#                     query.lonlat_box(north=lat_range[1], \
#                                      south=lat_range[0], \
#                                       east=lon_range[1],  \
#                                       west=lon_range[0]).vertical_level(10.0)
#                     query.variables("u-component_of_wind_height_above_ground",
#                                     "v-component_of_wind_height_above_ground")
#                     data = ncss.get_data(query)
#                     data = xr.open_dataset(NetCDF4DataStore(data))
#                     u = data['u-component_of_wind_height_above_ground']
#                     v = data['v-component_of_wind_height_above_ground']
#                     dataset.unit = u.units
#                     u = u.metpy.unit_array.squeeze()
#                     dataset.u[it,:,:] = np.array(u)
#                     v = v.metpy.unit_array.squeeze()
#                     dataset.v[it,:,:] = np.array(v)
                
#                 else:
                    
#                     # Other scalar variables                    
#     #                fac = 1.0                                     
#                     if param == "barometric_pressure":
#                         var_name = "Pressure_reduced_to_MSL_msl"
#                     elif param == "precipitation":    
#                         var_name = var_prcp
#     #                    fac = 1.0
#                     query = ncss.query()            
#                     query.lonlat_box(north=lat_range[1], \
#                                      south=lat_range[0], \
#                                      east=lon_range[1],  \
#                                      west=lon_range[0])
#                     query.variables(var_name)                    
#                     data = ncss.get_data(query)
#                     data = xr.open_dataset(NetCDF4DataStore(data))
#                     val          = data[var_name]
#                     dataset.unit = val.units
#                     val          = np.array(val.metpy.unit_array.squeeze())
#                     if param == "precipitation":
#                         # Data is stored either as 3-hourly (at 03h) or 6-hourly (at 06h) accumulated rainfall
#                         # For the first, just divide by 3 to get hourly precip
#                         # For the second, first subtract volume that fell in the first 3 hours
#                         if h==0 or h==6 or h==12 or h==18:
#                             val = val/3 # Convert to mm/h
#                         else:
#                             val = (val - 3*np.squeeze(dataset.val[it-1,:,:]))/3
#                     dataset.val[it,:,:]  = val

#             except:
                
#                 print("Could not download data")

#         # Write data to netcdf
#         time_string = time.strftime("%Y%m%d_%H%M")
#         file_name = prefix + "." + time_string + ".nc"
#         full_file_name = os.path.join(path, file_name)
#         ds = xr.Dataset()

             
#         okay = False
#         for ind, dataset in enumerate(datasets):

#             if dataset.quantity == "wind":

#                 uu = dataset.u[it,:,:]
#                 vv = dataset.v[it,:,:]
                
#                 if not np.any(np.isnan(uu)) and not np.any(np.isnan(vv)):
                    
#                     okay = True

#                     da = xr.DataArray(uu,
#                                           coords=[("lat", dataset.y),
#                                                   ("lon", dataset.x)])
#                     ds["wind_u"] = da
    
#                     da = xr.DataArray(vv,
#                                           coords=[("lat", dataset.y),
#                                                   ("lon", dataset.x)])
#                     ds["wind_v"] = da

#             else:

#                 val = dataset.val[it,:,:]

#                 if not np.any(np.isnan(val)):
                    
#                     okay = True
                    
#                     da = xr.DataArray(val,
#                                           coords=[("lat", dataset.y),
#                                                   ("lon", dataset.x)])
#                     ds[dataset.quantity] = da
#         if okay:
#             # Only write to file if there is any data
#             ds.to_netcdf(path=full_file_name)
    
#     return datasets

# # Helper function for finding proper time variable
# def find_time_var(var, time_basename='time'):
#     for coord_name in var.coords:
#         if coord_name.startswith(time_basename):
#             return var.coords[coord_name]
#     raise ValueError('No time variable found for ' + var.name)



# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""

import os
from datetime import datetime

import numpy as np
import pandas as pd
import xarray as xr
from pyproj import CRS


class Dataset:
    def __init__(self):
        self.quantity = None
        self.unit = None
        self.time = []
        self.x = None
        self.y = None
        self.crs = None
        self.val = None
        self.u = None
        self.v = None


def download(
    param_list, lon_range, lat_range, path, prefix, time_range=None, times=None
):
    if times is not None:
        requested_times = times
        time_range = [times[0], times[-1]]
    else:
        requested_times = (
            pd.date_range(start=time_range[0], end=time_range[1], freq="3H")
            .to_pydatetime()
            .tolist()
        )

    ntime = len(requested_times)

    datasets = []
    for param in param_list:
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)
        dataset.quantity = param
        datasets.append(dataset)

    icont = False
    # Get lat,lon
    for it, time in enumerate(requested_times):
        try:
            h = requested_times[it].hour
            month_string = requested_times[it].strftime("%Y%m")
            date_string = requested_times[it].strftime("%Y%m%d")

            toldnew = datetime(2020, 5, 15, 0, 0, 0, 0)
            cstr = "0000_000"

            if time < toldnew:
                # Analysis data before May 15th, 2020 stored in different url
                base_url = "https://www.ncei.noaa.gov/thredds/dodsC/model-gfs-g4-anl-files-old/"
                url = base_url + month_string + "/" + date_string + "/"
                name = "gfsanl_4_" + date_string + "_" + cstr + ".grb2"
            else:
                base_url = (
                    "https://www.ncei.noaa.gov/thredds/dodsC/model-gfs-g4-anl-files/"
                )
                url = base_url + month_string + "/" + date_string + "/"
                name = "gfs_4_" + date_string + "_" + cstr + ".grb2"

            # try:
            #     gfs   = TDSCatalog(url)
            # except:
            #     gfs   = []
            # okay = False
            # if gfs:
            #     for j, ds in enumerate(gfs.datasets):
            #         if ds==name:
            #             okay = True
            #             break
            # if not okay:
            #     # Try the next time
            #     continue

            ds = xr.open_dataset(url + name)  # or engine='pydap'

            lon_range[0] = np.mod(lon_range[0], 360.0)
            lon_range[1] = np.mod(lon_range[1], 360.0)

            i1 = np.where(ds.lat.to_numpy() > lat_range[1])[0][-1]
            i2 = np.where(ds.lat.to_numpy() < lat_range[0])[0][0]
            j1 = np.where(ds.lon.to_numpy() < lon_range[0])[0][-1]
            j2 = np.where(ds.lon.to_numpy() > lon_range[1])[0][0]

            # Latitude and longitude
            lat = ds.lat.to_numpy()[i1:i2]
            lon = ds.lon.to_numpy()[j1:j2]

            nrows = len(lat)
            ncols = len(lon)

            # ncss  = gfs.datasets[j].subset()
            # query = ncss.query()
            # query.lonlat_box(north=lat_range[1], \
            #                  south=lat_range[0], \
            #                   east=lon_range[1],  \
            #                   west=lon_range[0]).vertical_level(10.0)
            # query.variables("u-component_of_wind_height_above_ground",
            #                 "v-component_of_wind_height_above_ground")
            # data = ncss.get_data(query)
            # data = xr.open_dataset(NetCDF4DataStore(data))
            # lon = np.array(data['lon'])
            # lat = np.array(data['lat'])

            # nrows = len(lat)
            # ncols = len(lon)

            # Latitude and longitude found, so we can stop now
            icont = True
            break

        except Exception:
            # Try another time
            pass

    if not icont:
        # Could not find any data
        print("Could not find any data in requested range !")
        datasets = []
        return datasets

    # initialize matrices
    for dataset in datasets:
        dataset.x = lon
        dataset.y = lat
        if dataset.quantity == "wind":
            dataset.u = np.empty((ntime, nrows, ncols))
            dataset.u[:] = np.NaN
            dataset.v = np.empty((ntime, nrows, ncols))
            dataset.v[:] = np.NaN
        else:
            dataset.val = np.empty((ntime, nrows, ncols))
            dataset.val[:] = np.NaN

    for it, time in enumerate(requested_times):
        h = time.hour
        month_string = time.strftime("%Y%m")
        date_string = time.strftime("%Y%m%d")
        url = base_url + month_string + "/" + date_string + "/"

        # try:
        #     gfs   = TDSCatalog(url)
        # except:
        #     print("Could not fetch catalogue")
        #     continue

        if h == 0:
            cstr = "0000_000"
            crstr = "0000_003"
            var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
        elif h == 3:
            cstr = "0000_003"
            crstr = "0000_006"
            var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
        elif h == 6:
            cstr = "0600_000"
            crstr = "0600_003"
            var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
        elif h == 9:
            cstr = "0600_003"
            crstr = "0600_006"
            var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
        elif h == 12:
            cstr = "1200_000"
            crstr = "1200_003"
            var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
        elif h == 15:
            cstr = "1200_003"
            crstr = "1200_006"
            var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
        elif h == 18:
            cstr = "1800_000"
            crstr = "1800_003"
            var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
        elif h == 21:
            cstr = "1800_003"
            crstr = "1800_006"
            var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
        else:
            print(
                "ERROR: cycle and cycle_last in xml-file can only be specified rounded off on 3 hour values!"
            )

        # Loop through requested parameters
        for ind, param in enumerate(param_list):
            dataset = datasets[ind]
            dataset.time.append(time)

            if param == "precipitation":
                if time < toldnew:
                    name = "gfsanl_4_" + date_string + "_" + crstr + ".grb2"
                else:
                    name = "gfs_4_" + date_string + "_" + crstr + ".grb2"

            else:
                if time < toldnew:
                    name = "gfsanl_4_" + date_string + "_" + cstr + ".grb2"
                else:
                    name = "gfs_4_" + date_string + "_" + cstr + ".grb2"

            try:
                okay = False

                # for j, ds in enumerate(gfs.datasets):
                #     if ds==name:
                #         okay = True
                #         break

                # if not okay:
                #     # File not found, on to the next parameter
                #     print("Warning! " + name + " was not found on server")
                #     makezeros = True
                #     #continue

                print(name + " : " + param)

                for iattempt in range(10):
                    try:
                        ds = xr.open_dataset(url + name)
                        if iattempt > 0:
                            print("Success at attempt no " + int(iattempt + 1))
                        #                            makezeros = False
                        okay = True
                        break
                    except Exception:
                        # Try again
                        #                        makezeros = True
                        pass

                if okay:
                    if param == "wind":
                        u = ds["u-component_of_wind_height_above_ground"][
                            0, 0, i1:i2, j1:j2
                        ].to_numpy()
                        v = ds["v-component_of_wind_height_above_ground"][
                            0, 0, i1:i2, j1:j2
                        ].to_numpy()
                        dataset.u[it, :, :] = np.array(u)
                        dataset.v[it, :, :] = np.array(v)

                        # query = ncss.query()
                        # query.lonlat_box(north=lat_range[1], \
                        #                  south=lat_range[0], \
                        #                   east=lon_range[1],  \
                        #                   west=lon_range[0]).vertical_level(10.0)
                        # query.variables("u-component_of_wind_height_above_ground",
                        #                 "v-component_of_wind_height_above_ground")
                        # data = ncss.get_data(query)
                        # data = xr.open_dataset(NetCDF4DataStore(data))
                        # u = data['u-component_of_wind_height_above_ground']
                        # v = data['v-component_of_wind_height_above_ground']
                        # dataset.unit = u.units
                        # u = u.metpy.unit_array.squeeze()
                        # dataset.u[it,:,:] = np.array(u)
                        # v = v.metpy.unit_array.squeeze()
                        # dataset.v[it,:,:] = np.array(v)

                    else:
                        # Other scalar variables
                        #                fac = 1.0
                        if param == "barometric_pressure":
                            var_name = "Pressure_reduced_to_MSL_msl"
                        elif param == "precipitation":
                            var_name = var_prcp
                        val = ds[var_name][0, i1:i2, j1:j2].to_numpy()
                        # #                    fac = 1.0
                        #                 query = ncss.query()
                        #                 query.lonlat_box(north=lat_range[1], \
                        #                                  south=lat_range[0], \
                        #                                  east=lon_range[1],  \
                        #                                  west=lon_range[0])
                        #                 query.variables(var_name)
                        #                 data = ncss.get_data(query)
                        #                 data = xr.open_dataset(NetCDF4DataStore(data))
                        #                 val          = data[var_name]
                        #                 dataset.unit = val.units
                        #                 val          = np.array(val.metpy.unit_array.squeeze())
                        if param == "precipitation":
                            # Data is stored either as 3-hourly (at 03h) or 6-hourly (at 06h) accumulated rainfall
                            # For the first, just divide by 3 to get hourly precip
                            # For the second, first subtract volume that fell in the first 3 hours
                            if h == 0 or h == 6 or h == 12 or h == 18:
                                val = val / 3  # Convert to mm/h
                            else:
                                val = (
                                    val - 3 * np.squeeze(dataset.val[it - 1, :, :])
                                ) / 3
                        dataset.val[it, :, :] = val

                else:
                    print("Could not get data ...")

                # elif makezeros: # add zeros
                #      if param == "wind":
                #          dataset.u[:] = 0.0
                #          dataset.v[:] = 0.0
                #          print(param + " was not found on server ... --> using 0.0 m/s instead !!!")

                #      if param == "precipitation":
                #          dataset.val[:] = 0
                #          print(param + " was not found on server ... --> using 0.0 m/s instead !!!")

                #      if param == "barometric_pressure":
                #          dataset.val[:] = 102000.0
                #          print(param + " was not found on server ... --> using 102000.0 Pa instead !!!")

            except Exception:
                print("Could not download data")

        # Write data to netcdf
        time_string = time.strftime("%Y%m%d_%H%M")
        file_name = prefix + "." + time_string + ".nc"
        full_file_name = os.path.join(path, file_name)
        ds = xr.Dataset()

        okay = False
        for ind, dataset in enumerate(datasets):
            if dataset.quantity == "wind":
                uu = dataset.u[it, :, :]
                vv = dataset.v[it, :, :]

                if not np.any(np.isnan(uu)) and not np.any(np.isnan(vv)):
                    okay = True

                    da = xr.DataArray(
                        uu, coords=[("lat", dataset.y), ("lon", dataset.x)]
                    )
                    ds["wind_u"] = da

                    da = xr.DataArray(
                        vv, coords=[("lat", dataset.y), ("lon", dataset.x)]
                    )
                    ds["wind_v"] = da

                else:
                    print("NaNs found in wind. Skipping this time step.")

            else:
                val = dataset.val[it, :, :]

                if not np.any(np.isnan(val)):
                    okay = True

                    da = xr.DataArray(
                        val, coords=[("lat", dataset.y), ("lon", dataset.x)]
                    )
                    ds[dataset.quantity] = da

                else:
                    print(
                        "NaNs found in "
                        + dataset.quantity
                        + ". Skipping this time step."
                    )

        if okay:
            # Only write to file if there is any data
            ds.to_netcdf(path=full_file_name)
        else:
            print("No NetCDF file written for time " + time_string)

    return datasets


# Helper function for finding proper time variable
def find_time_var(var, time_basename="time"):
    for coord_name in var.coords:
        if coord_name.startswith(time_basename):
            return var.coords[coord_name]
    raise ValueError("No time variable found for " + var.name)


# Helper function for finding proper time variable
def find_height_var(var, time_basename="height"):
    for coord_name in var.coords:
        if coord_name.startswith(time_basename):
            return var.coords[coord_name]
    raise ValueError("No height variable found for " + var.name)



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 28 16:28:15 2021

@author: ormondt
"""

import json

import numpy as np
import yaml
from scipy.interpolate import RegularGridInterpolator


def interp2(x0, y0, z0, x1, y1, method="linear"):
    f = RegularGridInterpolator(
        (y0, x0), z0, bounds_error=False, fill_value=np.nan, method=method
    )
    # reshape x1 and y1
    if x1.ndim > 1:
        sz = x1.shape
        x1 = x1.reshape(sz[0] * sz[1])
        y1 = y1.reshape(sz[0] * sz[1])
        # interpolate
        z1 = f((y1, x1)).reshape(sz)
    else:
        z1 = f((y1, x1))

    return z1


def findreplace(file_name, str1, str2):
    # read input file
    fin = open(file_name, "rt")
    # read file contents to string
    data = fin.read()
    # replace all occurrences of the required string
    data = data.replace(str1, str2)
    # close the input file
    fin.close()
    # open the input file in write mode
    fin = open(file_name, "wt")
    # overwrite the input file with the resulting data
    fin.write(data)
    # close the file
    fin.close()


def read_json_js(file_name):
    # Read json javascript file (skipping the first line), and return json object
    fid = open(file_name, "r")
    lines = fid.readlines()
    fid.close()
    lines = lines[1:]
    jsn_string = ""
    for line in lines:
        # Strips the newline character
        jsn_string += line.strip()
    jsn = json.loads(jsn_string)
    return jsn


def write_json_js(file_name, jsn, first_line):
    # Writes json javascript file
    if type(jsn) == list:
        f = open(file_name, "w")
        f.write(first_line + "\n")
        f.write("[\n")
        for ix, x in enumerate(jsn):
            json_string = json.dumps(x)
            if ix < len(jsn) - 1:
                f.write(json_string + ",")
            else:
                f.write(json_string)
            f.write("\n")
        f.write("]\n")
        f.close()
    else:
        f = open(file_name, "w")
        f.write(first_line + "\n")
        json_string = json.dumps(jsn)
        f.write(json_string + "\n")
        f.close()


def write_csv_js(file_name, csv_string, first_line):
    # Writes json javascript file
    csv_string = csv_string.replace(chr(13), "")
    f = open(file_name, "w")
    f.write(first_line + "\n")
    f.write(csv_string)
    f.write("`;")
    f.close()


def rgb2hex(rgb):
    return "%02x%02x%02x" % rgb


def dict2yaml(file_name, dct, sort_keys=False):
    yaml_string = yaml.dump(dct, sort_keys=sort_keys)
    file = open(file_name, "w")
    file.write(yaml_string)
    file.close()


def yaml2dict(file_name):
    file = open(file_name, "r")
    dct = yaml.load(file, Loader=yaml.FullLoader)
    return dct



# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""

import os
# from siphon.catalog import TDSCatalog
# from xarray.backends import NetCDF4DataStore
import xarray as xr
import numpy as np
# from pyproj import CRS
# from metpy.units import units
from datetime import datetime

from .dataset import MeteoDataset

class MeteoDatasetGFSAnalysis0p50(MeteoDataset):

    # Inherit from MeteoDomain
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Set some source information
        self.source_name = "gfs_analysis_0p50"
        self.source_type = "analysis"
        self.source_delay = 6
        self.source_cycle_interval = 6
        self.source_time_interval = 3
    
    def download_analysis_times(self, requested_times, **kwargs):
        """Downloads COAMPS-TC forecast cycle for a given storm number and cycle time"""

        # ntime = len(requested_times)
        toldnew = datetime(2020, 5, 15, 0, 0, 0, 0)

        lon_range = self.lon_range
        lat_range = self.lat_range
        # GFS longitude is defined in degrees east (0 - 360)
        # If the lon_range is defined in degrees west (-180 - 180), we need to convert it
        londeg = "east"
        if lon_range[0] < 0:
            londeg = "west"
            lon_range[0] = 360.0 + lon_range[0]
            lon_range[1] = 360.0 + lon_range[1]
        # lon_range[0] = np.mod(lon_range[0], 360.0)
        # lon_range[1] = np.mod(lon_range[1], 360.0)

        icont = False

        # Get lat,lon
        for it, time in enumerate(requested_times):
            
            try:
                
                h            = requested_times[it].hour
                month_string = requested_times[it].strftime("%Y%m")
                date_string  = requested_times[it].strftime("%Y%m%d")
                
                # cstr  = "0000_000"
            
                if h==0:
                    cstr  = "0000_000"
                elif h==3:    
                    cstr  = "0000_003"
                elif h==6:    
                    cstr  = "0600_000"
                elif h==9:    
                    cstr  = "0600_003"
                elif h==12:    
                    cstr  = "1200_000"
                elif h==15:    
                    cstr  = "1200_003"
                elif h==18:    
                    cstr  = "1800_000"
                elif h==21:    
                    cstr  = "1800_003"
                
                if time<toldnew:
                    # Analysis data before May 15th, 2020 stored in different url
                    base_url = "https://www.ncei.noaa.gov/thredds/dodsC/model-gfs-g4-anl-files-old/"
                    url = base_url + month_string + "/" + date_string + "/"
                    name = "gfsanl_4_" + date_string + "_" + cstr + ".grb2"
                else:     
                    base_url = "https://www.ncei.noaa.gov/thredds/dodsC/model-gfs-g4-anl-files/"
                    url = base_url + month_string + "/" + date_string + "/"
                    name = "gfs_4_" + date_string + "_" + cstr + ".grb2"

                full_url = url + name    
                ds0 = xr.open_dataset(full_url)

                # Latitude and longitude

                # Data will be stored in ascending lat order !
                lon = ds0.lon.values[:]
                lat = ds0.lat.values[:]

                # Get lat, lon indices        

                j1 = np.where(lon<lon_range[0])[0]
                if len(j1) > 0:
                    j1 = j1[-1]
                else:
                    j1 = 0

                j2 = np.where(lon>lon_range[1])[0]
                if len(j2) > 0:
                    j2 = j2[0]
                else:
                    j2 = len(lon)

                i1 = np.where(lat>lat_range[1])[0]
                if len(i1) > 0:
                    i1 = i1[-1]
                else:
                    i1 = 0

                i2 = np.where(lat<lat_range[0])[0]
                if len(i2) > 0:
                    i2 = i2[0]
                else:
                    i2 = len(lat)

                if i2 <= i1 or j2 <= j1:
                    print("Error: cut-out is empty with given x_range and y_range")
                    return    

                # Latitude and longitude
                lat = lat[i1:i2]
                lon = lon[j1:j2]

                lat = np.flip(lat)

                if londeg == "west":
                    lon = lon - 360.0
        
                # Latitude and longitude found, so we can stop now
                icont = True
                ds0.close()

                break

            except Exception as e:
                print(e)                
                # Try another time
                print("Could not read " + full_url + " !")
        
        if not icont:
            # Could not find any data
            print("Could not find any data in requested range !")
            return
        
        r0 = np.zeros((np.size(lat), np.size(lon)))
        
        # Let's try to parallelize this
        for it, time in enumerate(requested_times):

            # Make new xarray dataset
            ds = xr.Dataset()
            ds["lon"] = xr.DataArray(lon, dims=("lon"))
            ds["lat"] = xr.DataArray(lat, dims=("lat"))

            h            = time.hour
            month_string = time.strftime("%Y%m")
            date_string  = time.strftime("%Y%m%d")
            url = base_url + month_string + "/" + date_string + "/"
            
            if h==0:
                cstr  = "0000_000"
                crstr = "0000_003"
                var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
            elif h==3:    
                cstr  = "0000_003"
                crstr = "0000_006"
                var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
            elif h==6:    
                cstr  = "0600_000"
                crstr = "0600_003"
                var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
            elif h==9:    
                cstr  = "0600_003"
                crstr = "0600_006"
                var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
            elif h==12:    
                cstr  = "1200_000"
                crstr = "1200_003"
                var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
            elif h==15:    
                cstr  = "1200_003"
                crstr = "1200_006"
                var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
            elif h==18:    
                cstr  = "1800_000"
                crstr = "1800_003"
                var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
            elif h==21:    
                cstr  = "1800_003"
                crstr = "1800_006"
                var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
            # else:
            #     print('ERROR: hours need to be round of to 3 hour values!')

            # Loop through requested parameters
            param_list = ["wind", "barometric_pressure", "precipitation"]

            for ind, param in enumerate(param_list):
                
                if param == "precipitation":
                    if time < toldnew:                
                        name = "gfsanl_4_" + date_string + "_" + crstr + ".grb2" 
                    else:
                        name = "gfs_4_" + date_string + "_" + crstr + ".grb2" 
                        
                else:    
                    if time < toldnew:                
                        name = "gfsanl_4_" + date_string + "_" + cstr + ".grb2" 
                    else:
                        name = "gfs_4_" + date_string + "_" + cstr + ".grb2" 

                try:
                                
                    okay = False
        
                    print(name + " : " + param)    
                    
                    for iattempt in range(10):    
                        try:
                            ds0 = xr.open_dataset(url + name)
                            if iattempt>0:
                                print("Success at attempt no " + int(iattempt + 1))
                            okay = True
                            break
                        except:
                            # Try again
                            pass

                    if okay:

                        if param == "wind":

                            u   = ds0["u-component_of_wind_height_above_ground"][0, 0, i1:i2, j1:j2].values
                            v   = ds0["v-component_of_wind_height_above_ground"][0, 0, i1:i2, j1:j2].values

                            u = np.flip(u, axis=0)
                            v = np.flip(v, axis=0)

                            ds["wind_u"] = xr.DataArray(u, dims=("lat", "lon"))
                            ds["wind_v"] = xr.DataArray(v, dims=("lat", "lon"))

                        else:

                            # Other scalar variables
                            if param == "barometric_pressure":
                                var_name = "Pressure_reduced_to_MSL_msl"
                            elif param == "precipitation":
                                var_name = var_prcp

                            val = ds0[var_name][0, i1:i2, j1:j2].values
                            val = np.flip(val, axis=0)

                            if param == "precipitation":
                                # Data is stored either as 3-hourly (at 03h) or 6-hourly (at 06h) accumulated rainfall
                                # For the first, just divide by 3 to get hourly precip
                                # For the second, first subtract volume that fell in the first 3 hours
                                if h==0 or h==6 or h==12 or h==18:
                                    val = val / 3 # Convert to mm/h
                                else:
                                    val = (val - 3 * np.squeeze(r0)) / 3

                                # Update r0
                                r0  = val    

                            ds[param] = xr.DataArray(val, dims=("lat", "lon"))

                    else:        
                        print("Could not get data ...")


                except:
                    
                    print("Could not download data")

            # Write to netcdf file
            ds.to_netcdf(path=os.path.join(self.path, self.name + "." + time.strftime("%Y%m%d_%H%M") + ".nc"))
            ds.close()



    # def download_time(time, lon, lat, base_url, toldnew):

    #     # Make new xarray dataset
    #     ds = xr.Dataset()
    #     ds["lon"] = xr.DataArray(lon, dims=("lon"))
    #     ds["lat"] = xr.DataArray(lat, dims=("lat"))

    #     h            = time.hour
    #     month_string = time.strftime("%Y%m")
    #     date_string  = time.strftime("%Y%m%d")
    #     url = base_url + month_string + "/" + date_string + "/"
        
    #     if h==0:
    #         cstr  = "0000_000"
    #         crstr = "0000_003"
    #         var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
    #     elif h==3:    
    #         cstr  = "0000_003"
    #         crstr = "0000_006"
    #         var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
    #     elif h==6:    
    #         cstr  = "0600_000"
    #         crstr = "0600_003"
    #         var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
    #     elif h==9:    
    #         cstr  = "0600_003"
    #         crstr = "0600_006"
    #         var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
    #     elif h==12:    
    #         cstr  = "1200_000"
    #         crstr = "1200_003"
    #         var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
    #     elif h==15:    
    #         cstr  = "1200_003"
    #         crstr = "1200_006"
    #         var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
    #     elif h==18:    
    #         cstr  = "1800_000"
    #         crstr = "1800_003"
    #         var_prcp = "Total_precipitation_surface_3_Hour_Accumulation"
    #     elif h==21:    
    #         cstr  = "1800_003"
    #         crstr = "1800_006"
    #         var_prcp = "Total_precipitation_surface_6_Hour_Accumulation"
    #     # else:
    #     #     print('ERROR: hours need to be round of to 3 hour values!')

    #     # Loop through requested parameters
    #     param_list = ["wind", "barometric_pressure", "precipitation"]

    #     for ind, param in enumerate(param_list):
            
    #         if param == "precipitation":
    #             if time < toldnew:                
    #                 name = "gfsanl_4_" + date_string + "_" + crstr + ".grb2" 
    #             else:
    #                 name = "gfs_4_" + date_string + "_" + crstr + ".grb2" 
                    
    #         else:    
    #             if time < toldnew:                
    #                 name = "gfsanl_4_" + date_string + "_" + cstr + ".grb2" 
    #             else:
    #                 name = "gfs_4_" + date_string + "_" + cstr + ".grb2" 

    #         try:
                            
    #             okay = False
    
    #             print(name + " : " + param)    
                
    #             for iattempt in range(10):    
    #                 try:
    #                     ds0 = xr.open_dataset(url + name)
    #                     if iattempt>0:
    #                         print("Success at attempt no " + int(iattempt + 1))
    #                     okay = True
    #                     break
    #                 except:
    #                     # Try again
    #                     pass

    #             if okay:

    #                 if param == "wind":

    #                     u   = ds0["u-component_of_wind_height_above_ground"][0, 0, i1:i2, j1:j2].values
    #                     v   = ds0["v-component_of_wind_height_above_ground"][0, 0, i1:i2, j1:j2].values

    #                     u = np.flip(u, axis=0)
    #                     v = np.flip(v, axis=0)

    #                     ds["wind_u"] = xr.DataArray(u, dims=("lat", "lon"))
    #                     ds["wind_v"] = xr.DataArray(v, dims=("lat", "lon"))

    #                 else:

    #                     # Other scalar variables
    #                     if param == "barometric_pressure":
    #                         var_name = "Pressure_reduced_to_MSL_msl"
    #                     elif param == "precipitation":
    #                         var_name = var_prcp

    #                     val = ds0[var_name][0, i1:i2, j1:j2].values
    #                     val = np.flip(val, axis=0)

    #                     if param == "precipitation":
    #                         # Data is stored either as 3-hourly (at 03h) or 6-hourly (at 06h) accumulated rainfall
    #                         # For the first, just divide by 3 to get hourly precip
    #                         # For the second, first subtract volume that fell in the first 3 hours
    #                         if h==0 or h==6 or h==12 or h==18:
    #                             val = val / 3 # Convert to mm/h
    #                         else:
    #                             val = (val - 3 * np.squeeze(r0)) / 3

    #                         # Update r0
    #                         r0  = val    

    #                     ds[param] = xr.DataArray(val, dims=("lat", "lon"))

    #             else:        
    #                 print("Could not get data ...")


    #         except:
                
    #             print("Could not download data")




# # Helper function for finding proper time variable
# def find_time_var(var, time_basename='time'):
#     for coord_name in var.coords:
#         if coord_name.startswith(time_basename):
#             return var.coords[coord_name]
#     raise ValueError('No time variable found for ' + var.name)

# # Helper function for finding proper time variable
# def find_height_var(var, time_basename='height'):
#     for coord_name in var.coords:
#         if coord_name.startswith(time_basename):
#             return var.coords[coord_name]
#     raise ValueError('No height variable found for ' + var.name)



# -*- coding: utf-8 -*-
from metget.metget_build import MetGetBuildRest
import requests
import tomli
import pandas as pd
import numpy as np
from pyproj import CRS
import datetime
from cht_meteo import gfs_forecast_0p25
from .coamps_utils import *

class Dataset():

    def __init__(self):

        self.quantity = None
        self.unit     = None
        self.time     = []
        self.x        = None
        self.y        = None
        self.crs      = None
        self.val      = None
        self.source   = []


def download(param_list, lon_range, lat_range, time_range, cycle_time, resolution=0.25, config_path=None):
    """Function to download coamps-tc forecasts using the metget tool api
    Right now resolution is hardcoded to 0.25 degrees but this can change"""

    fill_values = dict(wind=0.0, precipitation=0.0, barometric_pressure=102000.0)  # fill values to use in case of nan
    units = dict(wind='m/s', precipitation='kg.m-2.hour-1', barometric_pressure='Pa')  # units

    if not config_path:
        raise ValueError("Configuration data need to be provided")
    else:
        with open(config_path, mode="rb") as fp:
            config = tomli.load(fp)
            
    # Read data from the config file
    apikey = config["apikey"]
    endpoint = config["endpoint"]
    api_version = config["api_version"]
    if "tau" in config:
        tau = config["tau"]
    else:
        tau = 0
    if "override" in config:
        override = True
    else:
        override = False

    # Connect parameters names with coamps-tc names
    param_names = {'wind': 'wind_pressure', 'barometric_pressure': 'wind_pressure', 'precipitation': 'rain'}
    variables = list(np.unique([param_names[name] for name in param_list])) # ['wind_pressure', 'rain']

    # Get individual times requested
    # TODO how is time-step handled in cosmos?
    requested_times = pd.date_range(start=time_range[0],
                      end=time_range[1],
                      freq='3H').to_pydatetime().tolist()
    requested_times = [ti.replace(tzinfo=datetime.timezone.utc) for ti in requested_times]

    timestep = int((requested_times[1] - requested_times[0]).total_seconds())  # get time-step of requested time assuming that the time-step stays the same
    ntime = len(requested_times)

    # Get the available storms in the Coamps-TC forecast
    try:
        coamps_storms = check_coamps(apikey, endpoint, time_range[0])
    except:
        coamps_storms = {}
    
    storms = [name for name in coamps_storms.keys() if int(name.split("L")[0]) < 90]
    storms.sort()
    
    # Check if forecast exists for the selected dates and make a dataframe
    fr = pd.DataFrame(index=requested_times)
    fr["storms"] = None

    for i, ti in enumerate(requested_times):
        fr.loc[ti, "storms"] = []
        for j, storm_id in enumerate(storms):
            storm = coamps_storms[storm_id]                 
            ####### CHECK WHICH times to use for range ####################
            t1 = date_transform(storm['min_forecast_date'])
            t2 = date_transform(storm['max_forecast_date'])
            # t1 = date_transform(storm['latest_complete_forecast_start'])
            # t2 = date_transform(storm['latest_complete_forecast_end'])
            belongs = t1 <= ti <= t2
            if belongs:
                fr.loc[ti, "storms"].append(storm_id)

    # Check if there is a priority storm given
    if "priority_storm" in config:
        if config["priority_storm"] == "tc_vitals":
            # Get storm priority from noaa
            priority_storm = tc_vitals_storm()
        else:
            priority_storm = config["priority_storm"]
    else:
        priority_storm=None

    # Method to use when multiple storm are available without priority storm
    if "multi_storms_method" in config:
        method = config["multi_storms_method"]
    else:
        method = "last"

    fr["storm_id"] = None
    # Loop through time steps and choose a single storm to use 
    for i, ti in enumerate(requested_times):
        # # Check if for the chosen storm the time is during the skip time (tau) of the first forecast
        # for sn in fr.loc[ti, "storms"]:
        #     t0 = date_transform(coamps_storms[sn]["first_available_cycle"])
        #     t0_tau = t0 + datetime.timedelta(hours=tau)
        #     if ti <= t0_tau:
        #         fr.loc[ti, "storms"].remove(sn)
        storm_name = None # if there is no coamps storm make None so gfs is used
        if priority_storm and override:
            storm_name = priority_storm
        elif priority_storm in fr.loc[ti, "storms"]:  # if the priority storm is available that timestep use it
            storm_name = priority_storm
        else:
            # else used method defined
            if method == 'first':
                storm_name = fr.loc[ti, "storms"][0]
            elif method == 'last':
                storm_name = fr.loc[ti, "storms"][-1]

        # Assign storm
        fr.loc[ti, "storm_id"] = storm_name
        
    storm_log = np.array([v is not None for v in fr['storm_id'].values]) # check for which times there are forecasts from coamps-tc
    if all(~storm_log):  # If no coamps-tc data are available, gfs data are downloaded
        datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
        print("Could not find any coamps-tc data in requested time range! Using gfs instead")
        return datasets
    
    # Else if there are coamps-tc data see which storms (models) to use
    storms_id = np.unique(fr['storm_id'][storm_log])
    dss = {}  # prepare dictionary to save the datasets
    metadata = {}
    # Loop for the case there is more than one storm in requested times and get storm forecast
    for i, st in enumerate(storms_id):
        dss[st] = {}
        metadata[st] = {}
        if not override:
            storm = coamps_storms[st]
            t1 = datetime.datetime.strptime(storm['min_forecast_date'], "%Y-%m-%d %H:%M:%S").replace(tzinfo=datetime.timezone.utc)
            t2 = datetime.datetime.strptime(storm['max_forecast_date'], "%Y-%m-%d %H:%M:%S").replace(tzinfo=datetime.timezone.utc)
            # This parts ensures that requested data are within the model range
            t1 = max(requested_times[0], t1)
            t2 = min(requested_times[-1], t2)
        else:
            t1 = requested_times[0]
            t2 = requested_times[-1]
        if t1 == t2:  # scrubber tool cannot handle single time
            t1 = t1 - datetime.timedelta(hours=3)
        # Prepare metget domain
        domain = [f"coamps-{st}"] + [resolution] + [lon_range[0], lat_range[0], lon_range[1], lat_range[1]]

        for ii, var in enumerate(variables):
            # ...Building the request
            request_data = MetGetBuildRest.generate_request_json(
                start_date=t1.strftime("%Y%m%d %H%M%S"),
                end_date=t2.strftime("%Y%m%d %H%M%S"),
                format='hec-netcdf',
                data_type=var,
                time_step=timestep,
                domains=MetGetBuildRest.parse_command_line_domains([domain], tau),
                epsg=4326,
                filename=f"{st}_{var}",
                backfill=True,
                # nowcast=True,
                multiple_forecasts=True,
                # compression=True,
                save_json_request=False,
                # dry_run=True,
                strict=True,
            )
            client = MetGetBuildRest(endpoint, apikey, api_version)
            data_id, status_code = client.make_metget_request(request_data)
            urls, meta = client.download_metget_data(
                data_id,
                30, # sleep_time in seconds
                3, # max_wait in hours
                output_directory=None,
                return_only_url=True,
            )
            ds_list = []
            for url in urls:
                ds = get_da_from_url(url)
                ds_list.append(ds)
            if ds_list is not None:
                dss[st][var] = ds_list[0]
                metadata[st][var] = meta
            else:
                dss[st][var] = None

    # if no data could be downloaded (CHECK AGAIN WHAT SHOULD HAPPEN IN THIS CASE)
    if all([dss[k][kk] is None for k in list(dss.keys()) for kk in list(dss[k].keys())]):
        datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
        print("Could not find any coamps-tc data in requested time range! Using gfs instead")
        return datasets
    else: # Else if there are available data get lon lat info
        for key in dss[storms_id[0]].keys():
            if dss[storms_id[0]][key] is not None:
                data0 = dss[storms_id[0]][key]
                lon = np.array(data0['lon'])
                lat = np.array(data0['lat'])
                if lat[1] - lat[0] > 0:  # lat should be in descending order
                    lat = lat[::-1]
                    reverse = True
                else:
                    reverse = False
                lon = (lon + 360) % 360  # added this to be consistent with gfs
                nrows = len(lat)
                ncols = len(lon)

    # Prepare datasets
    datasets = []
    for param in param_list:
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)
        dataset.quantity = param
        dataset.x = lon
        dataset.y = lat
        dataset.time = pd.to_datetime([tt.replace(tzinfo=None) for tt in requested_times]).to_pydatetime()
        if dataset.quantity == "wind":
            dataset.u    = np.empty((ntime, nrows, ncols))
            dataset.u[:] = np.NaN
            dataset.v    = np.empty((ntime, nrows, ncols))
            dataset.v[:] = np.NaN
        else:
            dataset.val    = np.empty((ntime, nrows, ncols))
            dataset.val[:] = np.NaN
        datasets.append(dataset)


    for it, time_i in enumerate(requested_times):
        storm_id = fr['storm_id'][it]

        if storm_id is None:
            try:
                #  Currently data are used directly from gfs since grid is the same, but  resampling should be done in a new version
                datasets_gfs = gfs_forecast_0p25.download(param_list, lon_range, lat_range,
                                                          [time_i, time_i], cycle_time)
                for ind, param in enumerate(param_list):
                    if param == "wind":
                        datasets[ind].u[it, :, :] = datasets_gfs[ind].u
                        datasets[ind].v[it, :, :] = datasets_gfs[ind].v
                    else:
                        datasets[ind].val[it, :, :] = datasets_gfs[ind].val
                    datasets[ind].unit = datasets_gfs[ind].unit
                    datasets[ind].src.append('gfs')
                print("No coamps-tc forecast available for {} --> using gfs forecast instead !!!".format(time_i))

            except:
                print('gfs data not available')
            continue

        # Loop through requested parameters
        for ind, param in enumerate(param_list):
            dataset = datasets[ind]

            try:
                okay = False
                makezeros = False
                #  First check if there are coamps-tc data available for this parameter
                if dss[storm_id][param_names[param]] is not None:
                    okay = True
                    model_t = pd.to_datetime(dss[storm_id][param_names[param]].time.values).to_pydatetime()
                    model_t_ind = np.where(model_t == time_i.replace(tzinfo=None))[0][0]
                else:
                    makezeros = True

                if okay:
                    # Get metadata of input used
                    dataset.config = metadata[storm_id][param_names[param]]["input"]
                    inputs = [name for name in metadata[storm_id][param_names[param]]["input_files"][f"coamps-tc-{storm_id}"] if time_i.strftime("%Y%m%d%H%M") in name]
                    input_name = inputs[0].split(".")[2]
                    cycle_used = input_name.split("_")[2]
                    cycle_hour = input_name.split("_")[3]
                    dataset.source.append(f'coamps_tc_{storm_id}_{cycle_used}z_{cycle_hour}')
                    if param == "wind":
                        data = dss[storm_id][param_names[param]]
                        u = data['wind_u']
                        v = data['wind_v']
                        dataset.unit = u.units
                        
                        if np.any(np.isnan(u)) or np.any(np.isnan(v)):  # check if there are nans and fill them up
                            u = u.fillna(fill_values[param])
                            v = v.fillna(fill_values[param])
                        
                        u = u.metpy.unit_array.squeeze()
                        v = v.metpy.unit_array.squeeze()

                        if reverse:
                            dataset.u[it,:,:] = np.array(u[model_t_ind, ::-1, :])
                            dataset.v[it,:,:] = np.array(v[model_t_ind, ::-1, :])
                        else:
                            dataset.u[it,:,:] = np.array(u[model_t_ind, :, :])
                            dataset.v[it,:,:] = np.array(v[model_t_ind, :, :])
                    else:
                        if param == "barometric_pressure":
                            var_name = "mslp"
                        elif param == "precipitation":
                            var_name = 'rain' # Check when available in scrubber tool!
                        data = dss[storm_id][param_names[param]]
                        val          = data[var_name]

                        # Added this check to ensure that pressure is in Pa
                        if param == "barometric_pressure":
                            if val.units == 'mb':
                                val = val * 100
                                val.attrs['units'] = 'Pa'
                        dataset.unit = val.units

                        if param == "precipitation":  # Added this check to ensure that nan in precipitation are correctly interpreted
                            val = val.where(val>=0, np.nan)

                        if np.any(np.isnan(val)):  # check if there are nans and fill them up
                            val = val.fillna(fill_values[param])
                            
                        val = np.array(val.metpy.unit_array.squeeze())

                        if reverse:
                            dataset.val[it, :, :] = np.array(val[model_t_ind, ::-1, :])
                        else:
                            dataset.val[it, :, :] = np.array(val[model_t_ind, :, :])

                elif makezeros: # add zeros
                    try:
                        #  Currently data are used directly from gfs since grid is the same, but  resampling should be done in a new version
                        datasets_gfs = gfs_forecast_0p25.download(param_list, lon_range, lat_range,
                                                               [time_i, time_i], cycle_time)
                        if param == "wind":
                            dataset.u[it, :, :] = datasets_gfs[ind].u
                            dataset.v[it, :, :] = datasets_gfs[ind].v
                        else:
                            dataset.val[it, :, :]  = datasets_gfs[ind].val
                        dataset.unit = datasets_gfs[ind].unit
                        dataset.src.append('gfs')
                        print(param + " was not found on server for {} --> using gfs forecast data instead".format(time_i))
                    except:
                        if param == "wind":
                            dataset.u[:] = fill_values[param]
                            dataset.v[:] = fill_values[param]
                            dataset.unit = units[param]
                            dataset.src.append('default_value')
                            print(param + " was not found on server for {} --> using {} {} instead !!!".format(time_i, fill_values[param], units[param]))

                        else:
                            dataset.val[:] = fill_values[param]
                            dataset.unit = units[param]
                            dataset.src.append('default_value')
                            print(param + " was not found on server ... --> using {} {} instead !!!".format(fill_values[param], units[param]))

            except:
                print("Could not download data")

    return datasets



# -*- coding: utf-8 -*-
"""
@author:    Panos Athanasiou
Date:       04/07/2022
"""

import datetime
import json
import os
import time

import numpy as np
import pandas as pd
import requests
import xarray as xr
from pyproj import CRS


class Dataset:
    def __init__(self):
        self.quantity = None
        self.unit = None
        self.time = []
        self.x = None
        self.y = None
        self.crs = None
        self.val = None
        self.u = None
        self.v = None


def download(
    param_list,
    lon_range,
    lat_range,
    path,
    prefix,
    time_range=None,
    times=None,
    resolution=0.25,
):
    """Function to download coamps-tc forecasts using the scrubber tool functions
    Right now resolution is hardcoded to 0.25 degrees but this can change"""

    apikey = "GrPBTsclg71qgDvw4ojGf4mTwZmkbqNu6jcq7T9V"  # Deltares apikey hardcoded for the time being
    endpoint = "https://api.metget.zachcobell.com"

    # Connect parameters names with coamps-tc names
    param_names = {
        "wind": "wind_pressure",
        "barometric_pressure": "wind_pressure",
        "precipitation": "rain",
    }
    variables = list(
        np.unique([param_names[name] for name in param_list])
    )  # ['wind_pressure', 'rain']

    if times is not None:
        requested_times = times
        time_range = [times[0], times[-1]]
        # Add this step to make sure there are no issues when there already some netcdf for specific times
        # since the scrubber tool can only handle a time range and specific timestep
        requested_times = (
            pd.date_range(start=time_range[0], end=time_range[1], freq="3H")
            .to_pydatetime()
            .tolist()
        )
    else:
        requested_times = (
            pd.date_range(start=time_range[0], end=time_range[1], freq="3H")
            .to_pydatetime()
            .tolist()
        )

    timestep = int(
        (requested_times[1] - requested_times[0]).total_seconds()
    )  # get time-step of requested time assuming that the time-step stays the same
    ntime = len(requested_times)

    # Get storms from endpoint metadata
    storms = check_coamps(apikey, endpoint)

    # Check if forecast exists for the selected dates and make a dataframe
    fr = pd.DataFrame(index=requested_times)
    fr["storm_id"] = None
    for ti in requested_times:
        for j, storm in enumerate(storms):
            ####### CHECK WHICH times to use for range ####################
            t1 = datetime.datetime.strptime(
                storm["min_forecast_date"], "%Y-%m-%d %H:%M:%S"
            )
            t2 = datetime.datetime.strptime(
                storm["max_forecast_date"], "%Y-%m-%d %H:%M:%S"
            )
            # t1 = datetime.datetime.strptime(storm['latest_complete_forecast_start'], "%Y-%m-%d %H:%M:%S")
            # t2 = datetime.datetime.strptime(storm['latest_complete_forecast_end'], "%Y-%m-%d %H:%M:%S")
            if t1 <= ti <= t2:
                fr.loc[ti, "storm_id"] = storm[
                    "storm"
                ]  # latest available forecast for each time is saved

    # if no forecast available for any of the times return message
    storm_log = np.array([v is not None for v in fr["storm_id"].to_numpy()])
    if all(~storm_log):
        # Could not find any data
        print("Could not find any data in requested range !")
        datasets = []
        return datasets

    # else download period of interest with the latest available forecast for each time
    storms_id = np.unique(fr["storm_id"][storm_log])
    dss = {}  # prepare dictionary to save the datasets
    for i, st in enumerate(storms_id):
        dss[st] = {}
        for ii, var in enumerate(variables):
            ds = met_get(
                domain=[
                    [
                        "coamps-{}".format(st),
                        resolution,
                        lon_range[0],
                        lat_range[0],
                        lon_range[1],
                        lat_range[1],
                    ]
                ],  # list is needed here!
                start=requested_times[np.where(fr["storm_id"] == st)[0][0]].strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
                end=requested_times[np.where(fr["storm_id"] == st)[0][-1]].strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
                timestep=timestep,  # hardcoded for now! CHECK!
                variable=var,
                apikey=apikey,
                endpoint=endpoint,
                output="coamps_{}_{}".format(
                    st, var
                ),  # this is used for the online file creation
            )
            if ds is not None:
                dss[st][var] = ds[0]
            else:
                dss[st][var] = None

    # Prepare datasets
    datasets = []
    for param in param_list:
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)
        dataset.quantity = param
        datasets.append(dataset)

    # Get lat,lon
    for key in dss[storms_id[0]].keys():
        if dss[storms_id[0]][key] is not None:
            data0 = dss[storms_id[0]][key]
            lon = np.array(data0["lon"])
            lat = np.array(data0["lat"])
            if lat[1] - lat[0] > 0:  # lat should be in descending order
                lat = lat[::-1]
                reverse = True
            else:
                reverse = False

            nrows = len(lat)
            ncols = len(lon)

    # initialize matrices
    for dataset in datasets:
        dataset.x = lon
        dataset.y = lat
        if dataset.quantity == "wind":
            dataset.u = np.empty((ntime, nrows, ncols))
            dataset.u[:] = np.NaN
            dataset.v = np.empty((ntime, nrows, ncols))
            dataset.v[:] = np.NaN
        else:
            dataset.val = np.empty((ntime, nrows, ncols))
            dataset.val[:] = np.NaN

    for it, time_i in enumerate(requested_times):
        storm_id = fr["storm_id"][it]

        if storm_id is None:
            print("No forecast available for {}".format(time_i))
            continue

        model_ind = np.where(fr.index[fr["storm_id"] == storm_id] == time_i)[0][
            0
        ]  # this is the index of the time for the specific storm forecast

        # Loop through requested parameters
        for ind, param in enumerate(param_list):
            dataset = datasets[ind]
            dataset.time.append(time_i)

            try:
                okay = False
                makezeros = False

                if dss[storm_id][param_names[param]] is not None:
                    okay = True

                if not okay:
                    # File not found, on to the next parameter
                    print(
                        "Warning! "
                        + storm_id
                        + "_"
                        + param_names[param]
                        + " was not found on server"
                    )
                    makezeros = True
                    # continue
                print(storm_id + "_" + param_names[param] + " : " + param)

                if okay:
                    if param == "wind":
                        data = dss[storm_id][param_names[param]]
                        u = data["wind_u"]
                        v = data["wind_v"]
                        dataset.unit = u.units
                        u = u.metpy.unit_array.squeeze()
                        v = v.metpy.unit_array.squeeze()

                        if reverse:
                            dataset.u[it, :, :] = np.array(u[model_ind, ::-1, :])
                            dataset.v[it, :, :] = np.array(v[model_ind, ::-1, :])
                        else:
                            dataset.u[it, :, :] = np.array(u[model_ind, :, :])
                            dataset.v[it, :, :] = np.array(v[model_ind, :, :])
                    else:
                        # Other scalar variables
                        #                fac = 1.0
                        if param == "barometric_pressure":
                            var_name = "mslp"
                        elif param == "precipitation":
                            var_name = "rain"  # Check!
                        #                    fac = 1.0
                        data = dss[storm_id][param_names[param]]
                        val = data[var_name]
                        # Added this check to ensure that pressure is in Pa
                        if param == "barometric_pressure":
                            if val.units == "mb":
                                val = val * 100
                                val.attrs["units"] = "Pa"
                        dataset.unit = val.units
                        val = np.array(val.metpy.unit_array.squeeze())
                        if reverse:
                            dataset.val[it, :, :] = np.array(val[model_ind, ::-1, :])
                        else:
                            dataset.val[it, :, :] = np.array(val[model_ind, :, :])

                elif makezeros:  # add zeros
                    if param == "wind":
                        dataset.u[:] = 0
                        dataset.v[:] = 0
                        dataset.unit = "m/s"
                        print(
                            param
                            + " was not found on server ... --> using 0.0 m/s instead !!!"
                        )

                    if param == "precipitation":
                        dataset.val[:] = 0
                        dataset.unit = "m/s"
                        print(
                            param
                            + " was not found on server ... --> using 0.0 m/s instead !!!"
                        )

                    if param == "barometric_pressure":
                        dataset.val[:] = 102000.0
                        dataset.unit = "Pa"
                        print(
                            param
                            + " was not found on server ... --> using 102000.0 Pa instead !!!"
                        )
            except Exception:
                print("Could not download data")

        # Write data to netcdf
        time_string = time_i.strftime("%Y%m%d_%H%M")
        file_name = prefix + "." + time_string + ".nc"
        full_file_name = os.path.join(path, file_name)
        ds = xr.Dataset()

        okay = False
        for ind, dataset in enumerate(datasets):
            if dataset.quantity == "wind":
                uu = dataset.u[it, :, :]
                vv = dataset.v[it, :, :]

                # if not np.any(np.isnan(uu)) and not np.any(np.isnan(vv)):  # need to check this!
                if not np.all(np.isnan(uu)) and not np.all(np.isnan(vv)):
                    okay = True

                    da = xr.DataArray(
                        uu,
                        coords=[("lat", dataset.y), ("lon", dataset.x)],
                        attrs=dict(long_name="u component", unit=dataset.unit),
                    )
                    ds["wind_u"] = da

                    da = xr.DataArray(
                        vv,
                        coords=[("lat", dataset.y), ("lon", dataset.x)],
                        attrs=dict(long_name="v component", unit=dataset.unit),
                    )
                    ds["wind_v"] = da

            else:
                val = dataset.val[it, :, :]

                # if not np.any(np.isnan(val)): # need to check this!
                if not np.all(np.isnan(val)):
                    okay = True

                    da = xr.DataArray(
                        val,
                        coords=[("lat", dataset.y), ("lon", dataset.x)],
                        attrs=dict(long_name=dataset.quantity, unit=dataset.unit),
                    )

                    ds[dataset.quantity] = da
        if okay:
            # Save info on which storm forecast was used from the coamps-tc
            ds.attrs = {"coamps_storm_id": storm_id}
            # Only write to file if there is any data
            ds.to_netcdf(path=full_file_name)

    return datasets


def check_coamps(apikey, endpoint):
    """Read metadata of available forecasts from endpoint and returns a dictionary of the available storms"""
    headers = {"x-api-key": apikey}
    response_API = requests.get("{}/status".format(endpoint), headers=headers)
    data = json.loads(response_API.text)
    storms = data["body"]["data"]["metget"]["coamps-tc"]

    return storms


def parse_domain_data(domain_list: list, level) -> dict:
    import warnings

    AVAILABLE_MODELS = {
        "gfs": "gfs-ncep",
        "nam": "nam-ncep",
        "hwrf": "hwrf",
        "coamps": "coamps-tc",
    }

    model = domain_list[0]
    if "hwrf" in model:
        storm = model.split("-")[1]
        model = "hwrf"
        warnings.warn(
            "HWRF not fully supported yet. Use at your own risk.", RuntimeWarning
        )
    elif "coamps" in model:
        storm = model.split("-")[1]
        model = "coamps"
        warnings.warn(
            "COAMPS not fully supported yet. Use at your own risk.", RuntimeWarning
        )

    res = float(domain_list[1])
    x0 = float(domain_list[2])
    y0 = float(domain_list[3])
    x1 = float(domain_list[4])
    y1 = float(domain_list[5])

    if model not in AVAILABLE_MODELS.keys():
        raise RuntimeError("Specified model '" + model + "' is not available")

    xmax = max(x0, x1)
    xmin = min(x0, x1)
    ymax = max(y0, y1)
    ymin = min(y0, y1)
    res = abs(res)
    if res <= 0:
        raise RuntimeError("Specified model resolution is invalid")

    if model == "hwrf" or model == "coamps":
        return {
            "name": AVAILABLE_MODELS[model] + "-" + storm,
            "service": AVAILABLE_MODELS[model],
            "storm": storm,
            "x_init": xmin,
            "y_init": ymin,
            "x_end": xmax,
            "y_end": ymax,
            "di": res,
            "dj": res,
            "level": level,
        }
    else:
        return {
            "name": model,
            "service": AVAILABLE_MODELS[model],
            "x_init": xmin,
            "y_init": ymin,
            "x_end": xmax,
            "y_end": ymax,
            "di": res,
            "dj": res,
            "level": level,
        }


def make_metget_request(endpoint, apikey, request_json):
    headers = {"x-api-key": apikey}
    r = requests.post(endpoint + "/build", headers=headers, json=request_json)
    if r.status_code != 200:
        raise RuntimeError(
            "Request to MetGet was returned status code = " + str(r.status_code)
        )
    return_data = json.loads(r.text)
    data_id = return_data["body"]["request_id"]
    status_code = return_data["statusCode"]
    if status_code != 200:
        with open("metget.debug", "a") as f:
            f.write("[WARNING]: MetGet returned status code " + str(status_code) + "\n")
            f.write(str(return_data["body"]["error_text"]))
    return data_id, status_code


def download_metget_data(data_id, endpoint, apikey, sleeptime, max_wait):
    from datetime import datetime, timedelta

    # from netCDF4 import Dataset
    import xarray as xr

    # ...Status check
    headers = {"x-api-key": apikey}
    request_json = {"request": data_id}

    # ...Wait time
    end_time = datetime.utcnow() + timedelta(hours=max_wait)

    # ...Wait for request data to appear
    tries = 0
    data_ready = False
    status = None
    print("Waiting for request id: ", data_id, flush=True)
    while datetime.utcnow() <= end_time:
        tries += 1
        try:
            print(
                "["
                + datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC")
                + "]: Checking request status...(n="
                + str(tries)
                + "): ",
                flush=True,
                end="",
            )
            response = requests.post(
                endpoint + "/check", headers=headers, json=request_json
            )
            json_response = json.loads(response.text)
            status = json_response["body"]["status"]
            data_url = json_response["body"]["destination"]
            print(status, flush=True)
            if status == "completed":
                # ...Parse the return to get data
                data_ready = True
                flist_url = data_url + "/filelist.json"
                u = requests.get(flist_url)
                if u.status_code == 200:
                    return_data = json.loads(u.text)
                    # Turned this off to avoid saving this file
                    # with open("filelist.json", "w") as jsonfile:
                    #     jsonfile.write(
                    #         json.dumps(return_data, indent=2, sort_keys=True)
                    #     )
                    break
            elif status == "error":
                print("Request could not be completed")
                return
            else:
                time.sleep(sleeptime)
                continue
        except KeyboardInterrupt:
            print("[ERROR]: Process was ended by the user")
            raise

    # ...Download files
    if data_ready:
        file_list = return_data["output_files"]
        ds_list = []
        for f in file_list:
            time_stamp = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC")
            print("[{:s}]: Getting file: {:s}".format(time_stamp, f), flush=True)
            url = data_url + "/" + f
            # ds = Dataset('name', memory=requests.get(url).content)
            try:
                ds = xr.open_dataset(
                    url + "#mode=bytes"
                )  # Added this last part to allow opening with xarray
            except Exception:
                from netCDF4 import Dataset

                data = requests.get(url).content
                ds0 = Dataset("temp", memory=data)
                ds = xr.open_dataset(xr.backends.NetCDF4DataStore(ds0))
            ds_list.append(ds)

        return ds_list
    else:
        if status == "restore":
            print(
                "[WARNING]: Data for request "
                + data_id
                + " did not become ready before the max-wait time expired. You can rerun and ask for this request by id"
            )
        elif status == "running":
            print(
                "[WARNING]: Data for request "
                + data_id
                + " is still being constructed when the max-wait time expired. Please check on it later"
            )
        elif status == "queued":
            print(
                "[WARNING]: Data for request "
                + data_id
                + " is still queued. If this does not change soon, please contact an administrator"
            )
        else:
            print("[ERROR]: Data has not become available due to an unknown error")
        return


def met_get(**kwargs):
    """ "based on the get_metget_data.py script but instead of downloading a ntecdf files it returns an xr.dataset
    in memory"""
    import getpass
    import socket
    import types

    AVAILABLE_VARIABLES = {"wind_pressure", "rain", "temperature", "humidity", "ice"}
    AVAILABLE_FORMATS = {"ascii", "owi-ascii", "adcirc-netcdf", "hec-netcdf", "delft3d"}

    args = types.SimpleNamespace()

    for name in kwargs.keys():
        exec('args.{} = kwargs["{}"]'.format(name, name))

    optional = {
        "analysis": False,
        "multiple_forecasts": True,
        "format": "hec-netcdf",
        "variable": "wind_pressure",
        "check_interval": 30,
        "max_wait": 24,
        "strict": True,
        "backfile": True,
        "epsg": 4326,
        "dryrun": False,
        "request": None,
        "compression": False,
        "endpoint": None,
        "apikey": None,
    }

    for key in optional.keys():
        if key not in kwargs:
            exec('args.{} = optional["{}"]'.format(key, key))

    if not args.endpoint:
        if "METGET_ENDPOINT" not in os.environ:
            raise RuntimeError("No endpoint found.")
        else:
            endpoint = os.environ["METGET_ENDPOINT"]
    else:
        endpoint = args.endpoint

    if not args.apikey:
        if "METGET_API_KEY" not in os.environ:
            raise RuntimeError("No API key was found.")
        else:
            apikey = os.environ["METGET_API_KEY"]
    else:
        apikey = args.apikey

    # ...Check for required arguments
    if not args.request:
        if not args.start:
            print("[ERROR]: Must provide '--start'")
            exit(1)
        if not args.end:
            print("[ERROR]: Must provide '--end'")
            exit(1)
        if not args.timestep:
            print("[ERROR]: Must provide '--timestep'")
            exit(1)

        # ...Building the request
        domains = []
        idx = 0
        for d in args.domain:
            j = parse_domain_data(d, idx)
            domains.append(j)
            idx += 1

        if args.format not in AVAILABLE_FORMATS:
            print("ERROR: Invalid output format selected")
            exit(1)

        if args.format == "delft3d" or args.format == "hec-netcdf":
            if len(domains) > 1:
                print(
                    "[ERROR]: "
                    + args.format
                    + " does not support more than one domain."
                )
                exit(1)

        if args.variable not in AVAILABLE_VARIABLES:
            print("ERROR: Invalid variable selected")
            exit(1)

        request_from = getpass.getuser() + "." + socket.gethostname()
        request_data = {
            "version": "0.0.1",
            "creator": request_from,
            "background_pressure": 1013.0,
            "backfill": True,
            "nowcast": args.analysis,
            "multiple_forecasts": args.multiple_forecasts,
            "start_date": str(args.start),
            "end_date": str(args.end),
            "format": args.format,
            "data_type": args.variable,
            "time_step": args.timestep,
            "domains": domains,
            "compression": args.compression,
            "epsg": args.epsg,
            "filename": args.output,
        }
        if args.strict:
            request_data["strict"] = True
        if args.dryrun:
            request_data["dry_run"] = True

        data_id, status_code = make_metget_request(endpoint, apikey, request_data)
        if not args.dryrun and status_code == 200:
            ds_list = download_metget_data(
                data_id, endpoint, apikey, args.check_interval, args.max_wait
            )
        else:
            print(status_code)

    else:
        ds_list = download_metget_data(
            args.request, endpoint, apikey, args.check_interval, args.max_wait
        )
    return ds_list



# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""

import os
from siphon.catalog import TDSCatalog
from xarray.backends import NetCDF4DataStore
import xarray as xr
import pandas as pd
import numpy as np
from pyproj import CRS
from metpy.units import units

class Dataset():

    def __init__(self):

        self.quantity = None
        self.unit     = None
        self.time     = None
        self.x        = None
        self.y        = None
        self.crs      = None
        self.val      = None

def download(param_list, lon_range, lat_range, time_range, cycle_time,
             path, prefix):

    cycle_string = cycle_time.strftime("%Y%m%d_%H%M")

    base_url = "https://thredds.ucar.edu/thredds/catalog/grib/NCEP/GFS/Global_0p25deg/"
    url = base_url + "GFS_Global_0p25deg_" + cycle_string + ".grib2/catalog.xml"

    gfs   = TDSCatalog(url)
    ds    = list(gfs.datasets.values())[0]
    ncss  = ds.subset()
        
    datasets = {}

    # Loop through requested parameters     
    for param in param_list:
        
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)
 
        fac = 1.0

        if param == "wind_u":
            var_name = "u-component_of_wind_height_above_ground"
        elif param == "wind_v":    
            var_name = "v-component_of_wind_height_above_ground"
        elif param == "barometric_pressure":
            var_name = "Pressure_reduced_to_MSL_msl"
        elif param == "precipitation":    
            var_name = "Precipitation_rate_surface"
            fac = 3600.0
        
        dataset.quantity = param            
        query = ncss.query()            
                    
        if param == "wind_u" or param == "wind_v":
            query.lonlat_box(north=lat_range[1], \
                             south=lat_range[0], \
                             east=lon_range[1],  \
                             west=lon_range[0]).time_range(time_range[0], time_range[1]).vertical_level(10.0)
        else:
            query.lonlat_box(north=lat_range[1], \
                             south=lat_range[0], \
                             east=lon_range[1],  \
                             west=lon_range[0]).time_range(time_range[0], time_range[1])
            
        query.variables(var_name)
        data = ncss.get_data(query)
        data = xr.open_dataset(NetCDF4DataStore(data))

        dataset.x    = np.array(data['lon'])
        dataset.y    = np.array(data['lat'])

        val = data[var_name]
        time   = find_time_var(val)
        dataset.time = pd.to_datetime(time.data).to_pydatetime()
        dataset.unit = val.units
        val = val.metpy.unit_array.squeeze()
        dataset.val = np.array(val)*fac        
        
        datasets[param] = dataset

    # Write data to netcdf
    
    for it in range(len(datasets[param_list[0]].time)):

        time = datasets[param_list[0]].time[it]  
        time_string    = time.strftime("%Y%m%d_%H%M")
        file_name      = prefix + "." + time_string + ".nc"
        full_file_name = os.path.join(path, file_name)
        ds = xr.Dataset()
         
        okay = False
        
        for param in param_list:        
            
            val = datasets[param].val[it,:,:]    

            if not np.any(np.isnan(val)):                    
                okay = True                    
                da = xr.DataArray(val,
                                  coords=[("lat", datasets[param].y),
                                          ("lon", datasets[param].x)])
                ds[param] = da

        if okay:
            # Only write to file if there is any data
            ds.to_netcdf(path=full_file_name)
    
    return datasets

# Helper function for finding proper time variable
def find_time_var(var, time_basename='time'):
    for coord_name in var.coords:
        if coord_name.startswith(time_basename):
            return var.coords[coord_name]
    raise ValueError('No time variable found for ' + var.name)



# -*- coding: utf-8 -*-
import os
import datetime
import urllib
import tarfile as trf
import xarray as xr
import requests

from cht_utils import fileops as fo
from .dataset import MeteoDataset

class MeteoSubset:
    def __init__(self, name, moving):
        self.name = name
        self.moving = moving
        self.ds = xr.Dataset()

class MeteoDatasetCOAMPSTCForecastS3(MeteoDataset):

    # Inherit from MeteoDomain
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Set some source information
        self.source_name = "coamps_tc_forecast"
        self.source_type = "forecast"
        self.source_delay = 6
        self.source_cycle_interval = 6
        self.source_time_interval = 1

        # List subsets
        self.subset = []
        self.subset.append(MeteoSubset("d01", False))
        self.subset.append(MeteoSubset("d02", True))
        self.subset.append(MeteoSubset("d03", True))
    

    def download_forecast_cycle(self, **kwargs):
        """Downloads COAMPS-TC forecast cycle from S3 for a given storm number and cycle time"""

        if "cycle_time" in kwargs:
            cycle_time = kwargs["cycle_time"]
        else:
            # Throw error if cycle_time is not provided            
            print("Error: cycle_time not provided")
            return
        
        if "storm_number" in kwargs:
            storm_number = kwargs["storm_number"]
        else:
            # Throw error if storm_number is not provided            
            print("Error: storm_number not provided")
            return
        
        if "only_track" in kwargs:
            only_track = kwargs["only_track"]
        else:
            only_track = False

        # For COAMPS-TC we always get the entire dataset (lon_range, lat_range time_range are not used)

        base_url = "https://coamps-tc-data.s3.us-east-2.amazonaws.com"
        base_url = base_url + "/deterministic/realtime"

        # Get year from cycle_time
        year = cycle_time.year
        cycle_time_coamps = cycle_time.strftime("%Y%m%d%H")
        cycle_time_meteo  = cycle_time.strftime("%Y%m%d_%HZ")

        # Make folder for the forecast
        forecast_path = os.path.join(self.path, cycle_time_meteo) # Folder to store the forecast netcdf files
        os.makedirs(forecast_path, exist_ok=True) 

        # Start with the track file
        get_storm_track(forecast_path, year, storm_number, cycle_time_coamps)

        if only_track:
            # No need to download the gridded forecast, apparently
            return

        # Make folder for the temporary files
        tar_file_path = os.path.join(forecast_path, "_TMP")  # Temporary folder to store the tar file
        os.makedirs(tar_file_path, exist_ok=True)

        tar_file_name = storm_number + "_" + cycle_time_coamps + "_netcdf.tar"

        tar_file_full_path = os.path.join(tar_file_path, tar_file_name)

        url = base_url + "/" + str(year) + "/" + storm_number + "/" + cycle_time_coamps + "/" + tar_file_name

        # Download the tar file
        print(f"Downloading {url}")
        try:    
            urllib.request.urlretrieve(url, tar_file_full_path)
        except Exception as e:
            print(f"Error downloading {url}")
            print(e)
            # Remove the forecast folder
            fo.delete_folder(forecast_path)
            return

        # Extract the tar file
        with trf.open(tar_file_full_path, "r") as tar:
            for member in tar.getmembers():
                tar.extract(member, path=tar_file_path)

        # Convert all three resolutions to meteo format
        print("Converting COAMPS-TC netcdf files to meteo netcdf files")
        for subset in self.subset:
            res = subset.name
            # Get list of all the files in the _TMP/netcdf folder
            file_list = fo.list_files(os.path.join(tar_file_path, "netcdf", "*_" + res + "_*"))
            for file_name in file_list:
                # Read tau from file name
                tau = file_name.split("_")[-1].split(".")[0][3:]
                t = cycle_time + datetime.timedelta(hours=int(tau))
                tstr = t.strftime("%Y%m%d_%H%M")
                output_file = os.path.join(forecast_path, self.name + "." + res + "." + tstr + ".nc")
                convert_coamps_nc_to_meteo_nc(file_name, output_file)

        # Remove the temporary folder
        fo.delete_folder(tar_file_path)

        # Also try to download the track file
        track_file_name = storm_number + "_" + cycle_time_coamps + "_track.txt"

def convert_coamps_nc_to_meteo_nc(inpfile, outfile):
 
    # Open the COAMPS-TC netcdf file
    dsin = xr.open_dataset(inpfile)

    # Get the lon and lat
    lon = dsin["lon"].values[0, :] - 360.0
    lat = dsin["lat"].values[:, 0]

    # Create new dataset
    ds = xr.Dataset()
    # Add the lon and lat
    ds["lon"] = xr.DataArray(lon, dims=("lon"))
    ds["lat"] = xr.DataArray(lat, dims=("lat"))
    # Add the variables
    variables_coamps = ["uuwind", "vvwind", "slpres", "precip"]
    variables = ["wind_u", "wind_v", "barometric_pressure", "precipitation"]
    # units = ["m/s", "m/s", "hPa", "mm"]
    units = ["m/s", "m/s", "Pa", "mm/h"]
    for ivar, var in enumerate(variables_coamps):
        # Conversion factors
        fconv = 1.0
        if var == "slpres":
            fconv = 100.0
        if var in dsin:
            ds[variables[ivar]] = dsin[var] * fconv
            ds[variables[ivar]].attrs["units"] = units[ivar]

    # Write output file
    ds.to_netcdf(outfile)

    # Close the raw netcdf file
    dsin.close()

def get_storm_track(track_path:str, year:int, storm:str, cycle:str):
    """
    Retrieves the storm track data for a given year, storm, and cycle.

    Parameters:
    year (int): The year of the storm track data.
    storm (str): The name of the storm.
    cycle (str): The cycle of the storm track data.

    Returns:
    bytes: The content of the storm track data.

    """
    try:
        filename = os.path.join(track_path, f"TRK_COAMPS_CTCX_3_{cycle}_{storm}.trk")
        url = f"https://coamps-tc-data.s3.us-east-2.amazonaws.com/deterministic/realtime/{year}/{storm}/{cycle}/TRK_COAMPS_CTCX_3_{cycle}_{storm}"
        urllib.request.urlretrieve(url, filename)
    except Exception as e:
        print(f"Error downloading {url}")
        print(e)



# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""

import os

import numpy as np
import pandas as pd
import xarray as xr
from pyproj import CRS
from siphon.catalog import TDSCatalog
from xarray.backends import NetCDF4DataStore


class Dataset:
    def __init__(self):
        self.quantity = None
        self.unit = None
        self.time = None
        self.x = None
        self.y = None
        self.crs = None
        self.val = None


def download(param_list, lon_range, lat_range, time_range, cycle_time, path, prefix):
    cycle_string = cycle_time.strftime("%Y%m%d_%H%M")

    base_url = "https://thredds.ucar.edu/thredds/catalog/grib/NCEP/GFS/Global_0p25deg/"
    url = base_url + "GFS_Global_0p25deg_" + cycle_string + ".grib2/catalog.xml"

    gfs = TDSCatalog(url)
    ds = list(gfs.datasets.values())[0]
    ncss = ds.subset()

    datasets = {}

    # Loop through requested parameters
    for param in param_list:
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)

        fac = 1.0

        if param == "wind_u":
            var_name = "u-component_of_wind_height_above_ground"
        elif param == "wind_v":
            var_name = "v-component_of_wind_height_above_ground"
        elif param == "barometric_pressure":
            var_name = "Pressure_reduced_to_MSL_msl"
        elif param == "precipitation":
            var_name = "Precipitation_rate_surface"
            fac = 3600.0

        dataset.quantity = param
        query = ncss.query()

        if param == "wind_u" or param == "wind_v":
            query.lonlat_box(
                north=lat_range[1],
                south=lat_range[0],
                east=lon_range[1],
                west=lon_range[0],
            ).time_range(time_range[0], time_range[1]).vertical_level(10.0)
        else:
            query.lonlat_box(
                north=lat_range[1],
                south=lat_range[0],
                east=lon_range[1],
                west=lon_range[0],
            ).time_range(time_range[0], time_range[1])

        query.variables(var_name)
        data = ncss.get_data(query)
        data = xr.open_dataset(NetCDF4DataStore(data))

        dataset.x = np.array(data["lon"])
        dataset.y = np.array(data["lat"])

        val = data[var_name]
        time = find_time_var(val)
        dataset.time = pd.to_datetime(time.data).to_pydatetime()
        dataset.unit = val.units
        val = val.metpy.unit_array.squeeze()
        dataset.val = np.array(val) * fac

        datasets[param] = dataset

    # Write data to netcdf

    for it in range(len(datasets[param_list[0]].time)):
        time = datasets[param_list[0]].time[it]
        time_string = time.strftime("%Y%m%d_%H%M")
        file_name = prefix + "." + time_string + ".nc"
        full_file_name = os.path.join(path, file_name)
        ds = xr.Dataset()

        okay = False

        for param in param_list:
            val = datasets[param].val[it, :, :]

            if not np.any(np.isnan(val)):
                okay = True
                da = xr.DataArray(
                    val, coords=[("lat", datasets[param].y), ("lon", datasets[param].x)]
                )
                ds[param] = da

        if okay:
            # Only write to file if there is any data
            ds.to_netcdf(path=full_file_name)

    return datasets


# Helper function for finding proper time variable
def find_time_var(var, time_basename="time"):
    for coord_name in var.coords:
        if coord_name.startswith(time_basename):
            return var.coords[coord_name]
    raise ValueError("No time variable found for " + var.name)



import os
from datetime import datetime
from pathlib import Path

import psutil
from pyproj import CRS

from cht_meteo.meteo import MeteoGrid, MeteoSource


def test_download_meteo():
    path = Path(__file__).parent
    params = ["wind", "barometric_pressure", "precipitation"]
    lat = 32.77
    lon = -79.95

    # Download the actual datasets
    gfs_source = MeteoSource("gfs_anl_0p50", "gfs_anl_0p50_04", "hindcast", delay=None)

    # Create subset
    name = "gfs_anl_0p50_us_southeast"
    gfs_conus = MeteoGrid(
        name=name,
        source=gfs_source,
        parameters=params,
        path=path,
        x_range=[lon - 1, lon + 1],
        y_range=[lat - 1, lat + 1],
        crs=CRS.from_epsg(4326),
    )

    # Download and collect data
    t0 = datetime.strptime("20230101 000000", "%Y%m%d %H%M%S")
    t1 = datetime.strptime("20230101 020000", "%Y%m%d %H%M%S")
    time_range = [t0, t1]

    gfs_conus.download(time_range)
    assert path.joinpath("gfs_anl_0p50_us_southeast.20230101_0000.nc").is_file()

    gfs_conus.collect(time_range)

    assert gfs_conus.quantity[1].name == "barometric_pressure"
    assert gfs_conus.quantity[0].u.dtype == "float64"

    del gfs_conus, gfs_source

    delete_file(str(path.joinpath("gfs_anl_0p50_us_southeast.20230101_0000.nc")))


def delete_file(file_path):
    # Get the process ID of the process holding the file handle
    for proc in psutil.process_iter():
        try:
            for file in proc.open_files():
                if file.path == file_path:
                    pid = proc.pid
                    # Terminate the process
                    os.kill(pid, 9)
        except Exception:
            pass
    # Delete the file
    try:
        os.remove(file_path)
    except FileNotFoundError:
        print("File already removed")



import datetime
import pandas as pd
import requests
import xarray as xr

def date_transform(date):
    """
    Transforms a string representation of a date into a datetime object with UTC timezone.

    Args:
        date (str): A string representing a date in the format "%Y-%m-%d %H:%M:%S".

    Returns:
        datetime.datetime: A datetime object with the transformed date and UTC timezone.
    """
    return datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S").replace(tzinfo=datetime.timezone.utc)

def get_da_from_url(url):
    """
    Retrieves a dataset from a given URL and returns it as an xarray DataArray.

    Parameters:
    url (str): The URL of the dataset to retrieve.

    Returns:
    xr.DataArray: The dataset as an xarray DataArray.
    """
    from netCDF4 import Dataset as nc_Dataset
    data = requests.get(url).content
    ds0 = nc_Dataset('temp', memory=data)
    ds = xr.open_dataset(xr.backends.NetCDF4DataStore(ds0))
    # Alternative
    # ds = xr.open_dataset(url + '#mode=bytes') # Added this last part to allow opening with xarray
    return ds

def tc_vitals_storm():
    """Find the storm with the highest priority from NOAA"""
    try:
        tcvitals = requests.get('https://ftp.nhc.noaa.gov/atcf/com/tcvitals').text
        splits = [line.split() for line in tcvitals.split('\n')[:-1]] # last line is empty
        priority = pd.DataFrame(splits)
        priority = priority.iloc[[i for i in range(len(priority)) if 'L' in priority.iloc[i, 1]], :] # check only in atlantic basin storms
        pr_st_noaa = priority.iloc[0, 1] # define the top L storm as the priority one
        return pr_st_noaa
    except:
        return None


def check_coamps(apikey:str, endpoint:str, start:datetime=None, end:datetime=None) -> dict:
    """
    Read metadata of available forecasts from endpoint and returns a dictionary of the available storms.

    Parameters:
    - apikey (str): The API key for accessing the endpoint.
    - endpoint (str): The URL endpoint for retrieving the forecast data.
    - start (datetime): The start date for the forecast data.
    - end (datetime): The end date for the forecast data.

    Returns:
    - storms (dict): A dictionary containing the available storms and their metadata.
    """
    url = "{:s}/status?model={:s}".format(endpoint, "coamps")
    if start:
        url += "&start={:s}".format(start.strftime("%Y-%m-%d"))
    if end:
        url += "&end={:s}".format(end.strftime("%Y-%m-%d"))
    # ...Get the json from the endpoint
    response = requests.get(
        url, headers={"x-api-key": apikey}
    )
    data = response.json()["body"]
    
    if not data:
        return {}
    # TODO check if this is consistent with the latest endpoint
    if "data" in data:
        storms = data["data"]["metget"]["coamps-tc"]
    else:
        last_year = list(data.keys())[0]
        storms = data[last_year]
    return storms

def get_storm_track(year:int, storm:str, cycle:str):
    """
    Retrieves the storm track data for a given year, storm, and cycle.

    Parameters:
    year (int): The year of the storm track data.
    storm (str): The name of the storm.
    cycle (str): The cycle of the storm track data.

    Returns:
    bytes: The content of the storm track data.

    """
    url = f"https://coamps-tc-data.s3.us-east-2.amazonaws.com/deterministic/realtime/{year}/{storm}/{cycle}/TRK_COAMPS_CTCX_3_{cycle}_{storm}"
    response = requests.get(url)
    
    if response.status_code == 200:
        data = response.content
    else:
        data = None
    return data